"""
Geographically weighted PCA
===========================
Geographically Weighted Principal Component Analysis (GWPCA) is a spatial analysis method that identifies and visualizes local spatial patterns and relationships in multivariate datasets across various geographic areas. It operates by applying PCA within a moving window over a geographical region, which enables the extraction of local principal components that can differ across locations.

TIn this demonstration, we'll apply GWPCA to a dataset detailing the chemical compositions of soils from countries around the Baltic Sea [1]_. This example is inspired by a tutorial originally crafted and published by Chris Brunsdon [2]_.
The dataset comprises 10 variables (chemical elements) and spans 768 samples.
Here, each sample refers to a pair of latitude and longitude coordinates, representing specific sampling stations.

.. [1] Reimann, C. et al. Baltic soil survey: total concentrations of major and selected trace elements in arable soils from 10 countries around the Baltic Sea. Science of The Total Environment 257, 155–170 (2000).
.. [2] https://rpubs.com/chrisbrunsdon/99675



.. note:: The dataset we're using is found in the R package
    `mvoutlier <https://cran.r-project.org/web/packages/mvoutlier/mvoutlier.pdf>`_.
    To access it, we'll employ the Python package
    `rpy2 <https://rpy2.github.io/doc/latest/html/index.html>`_ which facilitates
    interaction with R packages from within Python.

.. note:: Presently, there's no support for ``xarray.Dataset`` lacking an explicit feature dimension.
    As a workaround, ``xarray.DataArray.to_array`` can be used to convert the ``Dataset`` to an ``DataArray``.

.. warning:: Bear in mind that GWPCA requires significant computational power.
    The ``xeofs`` implementation is optimized for CPU efficiency and is best suited
    for smaller to medium data sets. For more extensive datasets where parallel processing becomes essential,
    it's advisable to turn to the R package `GWmodel <https://cran.r-project.org/web/packages/GWmodel/GWmodel.pdf>`_.
    This package harnesses CUDA to enable GPU-accelerated GWPCA for optimized performance.


Let's import the necessary packages.
"""

# For the analysis
# For visualization
import matplotlib.pyplot as plt
import numpy as np

# For accessing R packages
import rpy2.robjects as ro
import seaborn as sns
import xarray as xr
from rpy2.robjects import pandas2ri
from rpy2.robjects.packages import importr

import xeofs as xe

# %%
# Next, we'll install the R package `mvoutlier <https://cran.r-project.org/web/packages/mvoutlier/mvoutlier.pdf>`_
# using the `rpy2 <https://rpy2.github.io/doc/latest/html/index.html>`_ package.

xr.set_options(display_expand_data=False)
utils = importr("utils")
utils.chooseCRANmirror(ind=1)
utils.install_packages("mvoutlier")

# %%
# Let's load the dataset and convert it into a ``pandas.DataFrame``.
# Alongside, we'll also load the background data that outlines the borders of countries
# in the Baltic Sea region. This will help us visually represent the GWPCA results.

ro.r(
    """
    require("mvoutlier")
    data(bsstop)
    Data <- bsstop[,1:14]
    background <- bss.background
    """
)
with (ro.default_converter + pandas2ri.converter).context():
    data_df = ro.conversion.get_conversion().rpy2py(ro.r["Data"])
    background_df = ro.conversion.get_conversion().rpy2py(ro.r["background"])
data_df.head()

# %%
# Since ``xeofs`` uses ``xarray``, we convert the data into an ``xarray.DataArray``.

data_df = data_df.rename(columns={"ID": "station"}).set_index("station")
data = data_df.to_xarray()
data = data.rename({"XCOO": "x", "YCOO": "y"})
data = data.set_index(station=("x", "y"))
data = data.drop_vars("CNo")
da = data.to_array(dim="element")
da

# %%
# Let's dive into the GWPCA. First, initialize a ``GWPCA`` instance and fit it to the data.
# The ``station`` dimension serves as our sample dimension, along which the local PCAs will be applied.
# Since these PCAs need to gauge distances to adjacent stations, we must specify
# a distance metric. Our station data includes coordinates in meters, so we'll
# choose the ``euclidean`` metric. If you have coordinates in degrees (like
# latitude and longitude), choose the ``haversine`` metric instead.
# We're also using a ``bisquare`` kernel with a bandwidth of 1000 km. Note that the
# bandwidth unit always follows input data (which is in meters here),
# except when using the ``haversine`` metric, which always gives distances in
# kilometers. Lastly, we'll standardize the input to ensure consistent scales
# for the chemical elements.

gwpca = xe.single.GWPCA(
    n_modes=5,
    standardize=True,
    metric="euclidean",
    kernel="bisquare",
    bandwidth=1000000.0,
)
gwpca.fit(da, "station")
gwpca.components()


# %%
# The ``components`` method returns the local principal components for each station. Note that the
# dimensionality of the returned array is ``[station, element, mode]``, so in practice we don't really have
# reduced the dimensionality of the data set. However, we can
# extract the largest locally weighted components for each station which tells us which chemical elements
# dominate the local PCAs.

llwc = gwpca.largest_locally_weighted_components()
llwc

# %%
# Let's visualize the spatial patterns of the chemical elements.
# As the stations are positioned on a irregular grid, we'll transform the
# ``llwc`` ``DataArray`` into a ``pandas.DataFrame``. After that, we can easily visualize
# it using the ``scatter`` method.
# For demonstation, we'll concentrate on the first mode:

llwc1_df = llwc.sel(mode=1).to_dataframe()

elements = da.element.values
n_elements = len(elements)
colors = np.arange(n_elements)
col_dict = {el: col for el, col in zip(elements, colors)}

llwc1_df["colors"] = llwc1_df["largest_locally_weighted_components"].map(col_dict)
cmap = sns.color_palette("tab10", n_colors=n_elements, as_cmap=True)


fig = plt.figure(figsize=(10, 10))
ax = fig.add_subplot(111)
background_df.plot.scatter(ax=ax, x="V1", y="V2", color=".3", marker=".", s=1)
s = ax.scatter(
    x=llwc1_df["x"],
    y=llwc1_df["y"],
    c=llwc1_df["colors"],
    ec="w",
    s=40,
    cmap=cmap,
    vmin=-0.5,
    vmax=n_elements - 0.5,
)
cbar = fig.colorbar(mappable=s, ax=ax, label="Largest locally weighted component")
cbar.set_ticks(colors)
cbar.set_ticklabels(elements)
ax.set_title("Largest locally weighted element", loc="left", weight=800)
plt.show()

# %%
# In the final step, let's examine the explained variance. Like standard PCA,
# this gives us insight into the variance explained by each mode. But with a
# local PCA for every station, the explained variance varies spatially. Notably,
# the first mode's explained variance differs across countries, ranging from
# roughly 40% to 70%.


exp_var_ratio = gwpca.explained_variance_ratio()
evr1_df = exp_var_ratio.sel(mode=1).to_dataframe()

fig = plt.figure(figsize=(10, 10))
ax = fig.add_subplot(111)
background_df.plot.scatter(ax=ax, x="V1", y="V2", color=".3", marker=".", s=1)
evr1_df.plot.scatter(
    ax=ax, x="x", y="y", c="explained_variance_ratio", vmin=0.4, vmax=0.7
)
ax.set_title("Fraction of locally explained variance", loc="left", weight=800)
plt.show()



import dask
import xarray as xr
from typing_extensions import Self

from ..utils.data_types import DataArray


class DataContainer(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._allow_compute = dict({k: True for k in self.keys()})

    def add(self, data: DataArray, name: str, allow_compute: bool = True) -> None:
        data.name = name
        super().__setitem__(name, data)
        self._allow_compute[name] = True if allow_compute else False

    def __setitem__(self, __key: str, __value: DataArray) -> None:
        super().__setitem__(__key, __value)
        self._allow_compute[__key] = self._allow_compute.get(__key, True)

    def __getitem__(self, __key: str) -> DataArray:
        try:
            return super().__getitem__(__key)
        except KeyError:
            raise KeyError(
                f"Cannot find data '{__key}'. Please fit the model first by calling .fit()."
            )

    def serialize(self) -> xr.DataTree:
        dt = xr.DataTree(name="data")
        for key, data in self.items():
            if not data.name:
                data.name = key
            dt[key] = xr.DataTree(data.to_dataset())
            dt[key].attrs = {key: "_is_node", "allow_compute": self._allow_compute[key]}

        return dt

    @classmethod
    def deserialize(cls, dt: xr.DataTree) -> Self:
        container = cls()
        for key, node in dt.items():
            container[key] = node[key]
            container._allow_compute[key] = node.attrs["allow_compute"]
        return container

    def compute(self, **kwargs):
        computed_data = {k: v for k, v in self.items() if self._allow_compute[k]}
        (computed_data,) = dask.compute(computed_data, **kwargs)
        for k, v in computed_data.items():
            self[k] = v

    def _validate_attrs_values(self, value):
        """Convert any boolean and None values to strings"""
        if isinstance(value, bool):
            return str(value)
        elif value is None:
            return "None"
        else:
            return value

    def _validate_attrs(self, attrs: dict) -> dict:
        """Convert any boolean and None values to strings"""
        for key, value in attrs.items():
            if isinstance(value, bool):
                attrs[key] = str(value)
            elif value is None:
                attrs[key] = "None"
            elif isinstance(value, list):
                attrs[key] = [self._validate_attrs_values(v) for v in value]

        return attrs

    def set_attrs(self, attrs: dict):
        attrs = self._validate_attrs(attrs)
        for key in self.keys():
            self[key].attrs = attrs



import pytest
import xarray as xr

from xeofs.cross import HilbertMCA


@pytest.fixture
def mca_model():
    return HilbertMCA(n_modes=3)


def test_initialization():
    mca = HilbertMCA(n_modes=1)
    assert mca is not None


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    assert hasattr(mca_model, "preprocessor1")
    assert hasattr(mca_model, "preprocessor2")
    assert hasattr(mca_model, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_squared_covariance_fraction(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    squared_covariance_fraction = mca_model.squared_covariance_fraction()
    assert isinstance(squared_covariance_fraction, xr.DataArray)
    assert (squared_covariance_fraction > 0).all()
    assert all(
        squared_covariance_fraction <= 1
    ), "Squared covariance fraction is greater than 1"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_covariance_fraction(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    cf = mca_model.covariance_fraction_CD95()
    assert isinstance(cf, xr.DataArray)
    assert all(cf <= 1), "Squared covariance fraction is greater than 1"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    components = mca_model.components()
    assert isinstance(components, tuple), "components is not a tuple"
    assert len(components) == 2, "components list does not have 2 elements"
    assert isinstance(components[0], xr.DataArray), "components[0] is not a DataArray"
    assert isinstance(components[1], xr.DataArray), "components[1] is not a DataArray"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    scores = mca_model.scores()
    assert isinstance(scores, tuple), "scores is not a tuple"
    assert len(scores) == 2, "scores list does not have 2 elements"
    assert isinstance(scores[0], xr.DataArray), "scores[0] is not a DataArray"
    assert isinstance(scores[1], xr.DataArray), "scores[1] is not a DataArray"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_amplitude(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    components = mca_model.components_amplitude()
    assert isinstance(components, tuple), "components is not a tuple"
    assert len(components) == 2, "components list does not have 2 elements"
    assert isinstance(components[0], xr.DataArray), "components[0] is not a DataArray"
    assert isinstance(components[1], xr.DataArray), "components[1] is not a DataArray"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_phase(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    components = mca_model.components_phase()
    assert isinstance(components, tuple), "components is not a tuple"
    assert len(components) == 2, "components list does not have 2 elements"
    assert isinstance(components[0], xr.DataArray), "components[0] is not a DataArray"
    assert isinstance(components[1], xr.DataArray), "components[1] is not a DataArray"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_amplitude(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    scores = mca_model.scores_amplitude()
    assert isinstance(scores, tuple), "scores is not a tuple"
    assert len(scores) == 2, "scores list does not have 2 elements"
    assert isinstance(scores[0], xr.DataArray), "scores[0] is not a DataArray"
    assert isinstance(scores[1], xr.DataArray), "scores[1] is not a DataArray"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_phase(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    scores = mca_model.scores_phase()
    assert isinstance(scores, tuple), "scores is not a tuple"
    assert len(scores) == 2, "scores list does not have 2 elements"
    assert isinstance(scores[0], xr.DataArray), "scores[0] is not a DataArray"
    assert isinstance(scores[1], xr.DataArray), "scores[1] is not a DataArray"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit_empty_data(dim):
    mca = HilbertMCA()
    with pytest.raises(ValueError):
        mca.fit(xr.DataArray(), xr.DataArray(), dim)


@pytest.mark.parametrize(
    "dim",
    [
        ("invalid_dim"),
    ],
)
def test_fit_invalid_dims(mca_model, mock_data_array, dim):
    with pytest.raises(ValueError):
        mca_model.fit(mock_data_array, mock_data_array, dim)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_transform_not_implemented(mca_model, mock_data_array, dim):
    with pytest.raises(NotImplementedError):
        mca_model.transform(mock_data_array, mock_data_array)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit_with_dataset(mca_model, mock_dataset, dim):
    mca_model.fit(mock_dataset, mock_dataset, dim)
    assert hasattr(mca_model, "preprocessor1")
    assert hasattr(mca_model, "preprocessor2")
    assert hasattr(mca_model, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit_with_datalist(mca_model, mock_data_array_list, dim):
    mca_model.fit(mock_data_array_list, mock_data_array_list, dim)
    assert hasattr(mca_model, "preprocessor1")
    assert hasattr(mca_model, "preprocessor2")
    assert hasattr(mca_model, "data")



from typing import Sequence

import numpy as np

from .cpcca import CPCCA, ComplexCPCCA, HilbertCPCCA


class RDA(CPCCA):
    """Redundancy Analysis (RDA).

    RDA seeks to find paris of coupled patterns that maximize the predictand
    variance [1]_ [2]_ .

    This method solves the following optimization problem:

        :math:`\\max_{q_x, q_y} \\left( q_x^T X^T Y q_y \\right)`

    subject to the constraints:

        :math:`q_x^T (X^TX) q_x = 1, \\quad q_y^T q_y = 1`

    where :math:`X` and :math:`Y` are the input data matrices and :math:`q_x`
    and :math:`q_y` are the corresponding pattern vectors.

    Parameters
    ----------
    n_modes : int, default=2
        Number of modes to calculate.
    standardize : Squence[bool] | bool, default=False
        Whether to standardize the input data. Generally not recommended as
        standardization can be managed by the degree of whitening.
    use_coslat : Sequence[bool] | bool, default=False
        For data on a longitude-latitude grid, whether to correct for varying
        grid cell areas towards the poles by scaling each grid point with the
        square root of the cosine of its latitude.
    use_pca : Sequence[bool] | bool, default=False
        Whether to preprocess each field individually by reducing dimensionality
        through PCA. The cross-covariance matrix is then computed in the reduced
        principal component space.
    n_pca_modes : Sequence[int | float | str] | int | float | str, default=0.999
        Number of modes to retain during PCA preprocessing step. If int,
        specifies the exact number of modes; if float, specifies the fraction of
        variance to retain; if "all", all modes are retained.
    pca_init_rank_reduction : Sequence[float] | float, default=0.3
        Relevant when `use_pca=True` and `n_pca_modes` is a float. Specifies the
        initial fraction of rank reduction for faster PCA computation via
        randomized SVD.
    check_nans : Sequence[bool] | bool, default=True
        Whether to check for NaNs in the input data. Set to False for lazy model
        evaluation.
    compute : bool, default=True
        Whether to compute the model elements eagerly. If True, the following
        are computed sequentially: preprocessor scaler, optional NaN checks, SVD
        decomposition, scores, and components.
    random_state : numpy.random.Generator | int | None, default=None
        Seed for the random number generator.
    sample_name : str, default="sample"
        Name for the new sample dimension.
    feature_name : Sequence[str] | str, default="feature"
        Name for the new feature dimension.
    solver : {"auto", "full", "randomized"}
        Solver to use for the SVD computation.
    solver_kwargs : dict, default={}
        Additional keyword arguments passed to the SVD solver function.


    References
    ----------
    .. [1] 1. Storch, H. von & Zwiers, F. W. Statistical Analysis in Climate
        Research. (Cambridge University Press (Virtual Publishing), 2003).
    .. [2] Wilks, D. S. Statistical Methods in the Atmospheric Sciences.
        (Academic Press, 2019).
        doi:https://doi.org/10.1016/B978-0-12-815823-4.00011-0.

    Examples
    --------

    Perform RDA on two datasets on a regular longitude-latitude grid:

    >>> model = RDA(n_modes=5, use_coslat=True)
    >>> model.fit(X, Y, dim="time")

    """

    def __init__(
        self,
        n_modes: int = 2,
        standardize: Sequence[bool] | bool = False,
        use_coslat: Sequence[bool] | bool = False,
        check_nans: Sequence[bool] | bool = True,
        use_pca: Sequence[bool] | bool = True,
        n_pca_modes: Sequence[float | int | str] | float | int | str = 0.999,
        pca_init_rank_reduction: Sequence[float] | float = 0.3,
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: Sequence[str] | str = "feature",
        solver: str = "auto",
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
    ):
        CPCCA.__init__(
            self,
            n_modes=n_modes,
            alpha=[0.0, 1.0],
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            use_pca=use_pca,
            n_pca_modes=n_pca_modes,
            pca_init_rank_reduction=pca_init_rank_reduction,
            compute=compute,
            sample_name=sample_name,
            feature_name=feature_name,
            solver=solver,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
        )
        self.attrs.update({"model": "Redundancy Analysis"})
        # Renove alpha from the inherited CPCCA serialization params because it is hard-coded for RDA
        self._params.pop("alpha")


class ComplexRDA(ComplexCPCCA, RDA):
    """Complex RDA.

    RDA applied to a complex-valued field obtained from a pair of variables such
    as the zonal and meridional components, :math:`U` and :math:`V`, of the wind
    field. Complex RDA analysis then maximizes the correlation between
    two datasets of the form

    .. math::
        Z_x = U_x + iV_x

    and

    .. math::
        Z_y = U_y + iV_y

    into a set of complex-valued components and PC scores.


    Parameters
    ----------
    n_modes : int, default=2
        Number of modes to calculate.
    standardize : Squence[bool] | bool, default=False
        Whether to standardize the input data. Generally not recommended as
        standardization can be managed by the degree of whitening.
    use_coslat : Sequence[bool] | bool, default=False
        For data on a longitude-latitude grid, whether to correct for varying
        grid cell areas towards the poles by scaling each grid point with the
        square root of the cosine of its latitude.
    use_pca : Sequence[bool] | bool, default=False
        Whether to preprocess each field individually by reducing dimensionality
        through PCA. The cross-covariance matrix is then computed in the reduced
        principal component space.
    n_pca_modes : Sequence[int | float | str] | int | float | str, default=0.999
        Number of modes to retain during PCA preprocessing step. If int,
        specifies the exact number of modes; if float, specifies the fraction of
        variance to retain; if "all", all modes are retained.
    pca_init_rank_reduction : Sequence[float] | float, default=0.3
        Relevant when `use_pca=True` and `n_pca_modes` is a float. Specifies the
        initial fraction of rank reduction for faster PCA computation via
        randomized SVD.
    check_nans : Sequence[bool] | bool, default=True
        Whether to check for NaNs in the input data. Set to False for lazy model
        evaluation.
    compute : bool, default=True
        Whether to compute the model elements eagerly. If True, the following
        are computed sequentially: preprocessor scaler, optional NaN checks, SVD
        decomposition, scores, and components.
    random_state : numpy.random.Generator | int | None, default=None
        Seed for the random number generator.
    sample_name : str, default="sample"
        Name for the new sample dimension.
    feature_name : Sequence[str] | str, default="feature"
        Name for the new feature dimension.
    solver : {"auto", "full", "randomized"}
        Solver to use for the SVD computation.
    solver_kwargs : dict, default={}
        Additional keyword arguments passed to the SVD solver function.

    Examples
    --------

    With two DataArrays `u_i` and `v_i` representing the zonal and meridional
    components of the wind field for two different regions :math:`x` and
    :math:`y`, construct

    >>> X = u_x + 1j * v_x
    >>> Y = u_y + 1j * v_y

    and fit the Complex RDA model:

    >>> model = ComplexRDA(n_modes=5)
    >>> model.fit(X, Y, "time")


    """

    def __init__(
        self,
        n_modes: int = 2,
        standardize: Sequence[bool] | bool = False,
        use_coslat: Sequence[bool] | bool = False,
        check_nans: Sequence[bool] | bool = True,
        use_pca: Sequence[bool] | bool = True,
        n_pca_modes: Sequence[float | int | str] | float | int | str = 0.999,
        pca_init_rank_reduction: Sequence[float] | float = 0.3,
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: Sequence[str] | str = "feature",
        solver: str = "auto",
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
    ):
        ComplexCPCCA.__init__(
            self,
            n_modes=n_modes,
            alpha=[0.0, 1.0],
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            use_pca=use_pca,
            n_pca_modes=n_pca_modes,
            pca_init_rank_reduction=pca_init_rank_reduction,
            compute=compute,
            sample_name=sample_name,
            feature_name=feature_name,
            solver=solver,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
        )
        self.attrs.update({"model": "Complex RDA"})
        # Renove alpha from the inherited CPCCA serialization params because it is hard-coded for RDA
        self._params.pop("alpha")


class HilbertRDA(HilbertCPCCA, ComplexRDA):
    """Hilbert RDA.

    Hilbert RDA  extends RDA by examining amplitude-phase relationships. It
    augments the input data with its Hilbert transform, creating a
    complex-valued field.

    This method solves the following optimization problem:

        :math:`\\max_{q_x, q_y} \\left( q_x^H X^H Y q_y \\right)`

    subject to the constraints:

        :math:`q_x^H (X^HX) q_x = 1, \\quad q_y^H q_y = 1`

    where :math:`H` denotes the conjugate transpose and :math:`X` and :math:`Y`
    are the augmented data matrices.

    An optional padding with exponentially decaying values can be applied prior
    to the Hilbert transform in order to mitigate the impact of spectral
    leakage.


    Parameters
    ----------
    n_modes : int, default=2
        Number of modes to calculate.
    padding : Sequence[str] | str | None, default="exp"
        Padding method for the Hilbert transform. Available options are: - None:
        no padding - "exp": exponential decay
    decay_factor : Sequence[float] | float, default=0.2
        Decay factor for the exponential padding.
    standardize : Squence[bool] | bool, default=False
        Whether to standardize the input data. Generally not recommended as
        standardization can be managed by the degree of whitening.
    use_coslat : Sequence[bool] | bool, default=False
        For data on a longitude-latitude grid, whether to correct for varying
        grid cell areas towards the poles by scaling each grid point with the
        square root of the cosine of its latitude.
    use_pca : Sequence[bool] | bool, default=False
        Whether to preprocess each field individually by reducing dimensionality
        through PCA. The cross-covariance matrix is computed in the reduced
        principal component space.
    n_pca_modes : Sequence[int | float | str] | int | float | str, default=0.999
        Number of modes to retain during PCA preprocessing step. If int,
        specifies the exact number of modes; if float, specifies the fraction of
        variance to retain; if "all", all modes are retained.
    pca_init_rank_reduction : Sequence[float] | float, default=0.3
        Relevant when `use_pca=True` and `n_pca_modes` is a float. Specifies the
        initial fraction of rank reduction for faster PCA computation via
        randomized SVD.
    check_nans : Sequence[bool] | bool, default=True
        Whether to check for NaNs in the input data. Set to False for lazy model
        evaluation.
    compute : bool, default=True
        Whether to compute the model elements eagerly. If True, the following
        are computed sequentially: preprocessor scaler, optional NaN checks, SVD
        decomposition, scores, and components.
    random_state : numpy.random.Generator | int | None, default=None
        Seed for the random number generator.
    sample_name : str, default="sample"
        Name for the new sample dimension.
    feature_name : Sequence[str] | str, default="feature"
        Name for the new feature dimension.
    solver : {"auto", "full", "randomized"}
        Solver to use for the SVD computation.
    solver_kwargs : dict, default={}
        Additional keyword arguments passed to the SVD solver function.



    Examples
    --------
    >>> model = HilbertRDA(n_modes=5)
    >>> model.fit(X, Y, "time")

    """

    def __init__(
        self,
        n_modes: int = 2,
        padding: Sequence[str] | str | None = "exp",
        decay_factor: Sequence[float] | float = 0.2,
        standardize: Sequence[bool] | bool = False,
        use_coslat: Sequence[bool] | bool = False,
        check_nans: Sequence[bool] | bool = True,
        use_pca: Sequence[bool] | bool = True,
        n_pca_modes: Sequence[float | int | str] | float | int | str = 0.999,
        pca_init_rank_reduction: Sequence[float] | float = 0.3,
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: Sequence[str] | str = "feature",
        solver: str = "auto",
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
    ):
        HilbertCPCCA.__init__(
            self,
            n_modes=n_modes,
            alpha=[0.0, 1.0],
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            use_pca=use_pca,
            n_pca_modes=n_pca_modes,
            pca_init_rank_reduction=pca_init_rank_reduction,
            compute=compute,
            sample_name=sample_name,
            feature_name=feature_name,
            solver=solver,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
            padding=padding,
            decay_factor=decay_factor,
        )
        self.attrs.update({"model": "Hilbert RDA"})
        # Renove alpha from the inherited CPCCA serialization params because it is hard-coded for RDA
        self._params.pop("alpha")



import numpy as np
import xarray as xr
from typing_extensions import Self

from ..linalg.decomposer import Decomposer
from ..utils.data_types import DataArray, DataObject
from ..utils.sanity_checks import assert_not_complex
from .base_model_single_set import BaseModelSingleSet
from .eof import EOF


class OPA(BaseModelSingleSet):
    """Optimal Persistence Analysis.

    Optimal Persistence Analysis (OPA) [1]_ [2]_ identifies the patterns with the
    largest decorrelation time in a time-varying field, known as optimal
    persistence patterns or optimally persistent patterns (OPP).

    Parameters
    ----------
    n_modes : int
        Number of optimal persistence patterns (OPP) to be computed.
    tau_max : int
        Maximum time lag for the computation of the covariance matrix.
    center : bool, default=True
        Whether to center the input data.
    standardize : bool, default=False
        Whether to standardize the input data.
    use_coslat : bool, default=False
        Whether to use cosine of latitude for scaling.
    n_pca_modes : int
        Number of modes to be computed in the pre-processing step using EOF.
    compute : bool, default=True
        Whether to compute elements of the model eagerly, or to defer computation.
        If True, four pieces of the fit will be computed sequentially: 1) the
        preprocessor scaler, 2) optional NaN checks, 3) SVD decomposition, 4) scores
        and components.
    sample_name : str, default="sample"
        Name of the sample dimension.
    feature_name : str, default="feature"
        Name of the feature dimension.
    solver : {"auto", "full", "randomized"}, default="auto"
        Solver to use for the SVD computation.
    solver_kwargs : dict, default={}
        Additional keyword arguments to pass to the solver.

    References
    ----------
    .. [1] DelSole, T. Optimally Persistent Patterns in Time-Varying Fields. Journal of the Atmospheric Sciences 58, 1341–1356 (2001).
    .. [2] DelSole, T. Low-Frequency Variations of Surface Temperature in Observations and Simulations. Journal of Climate 19, 4487–4507 (2006).

    Examples
    --------
    >>> from xeofs.single import OPA
    >>> model = OPA(n_modes=10, tau_max=50, n_pca_modes=100)
    >>> model.fit(X, dim=("time"))

    Retrieve the optimally persistent patterns (OPP) and their time series:

    >>> opp = model.components()
    >>> opp_ts = model.scores()

    Retrieve the decorrelation time of the OPPs:

    >>> decorrelation_time = model.decorrelation_time()
    """

    def __init__(
        self,
        n_modes: int,
        tau_max: int,
        center: bool = True,
        standardize: bool = False,
        use_coslat: bool = False,
        check_nans: bool = True,
        n_pca_modes: int = 100,
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: str = "feature",
        solver: str = "auto",
        random_state: int | None = None,
        solver_kwargs: dict = {},
    ):
        if n_modes > n_pca_modes:
            raise ValueError(
                f"n_modes must be smaller or equal to n_pca_modes (n_modes={n_modes}, n_pca_modes={n_pca_modes})"
            )
        super().__init__(
            n_modes=n_modes,
            center=center,
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            compute=compute,
            sample_name=sample_name,
            feature_name=feature_name,
            solver=solver,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
        )
        self.attrs.update({"model": "OPA"})
        self._params.update({"tau_max": tau_max, "n_pca_modes": n_pca_modes})

    def _Ctau(self, X, tau: int) -> DataArray:
        """Compute the time-lage covariance matrix C(tau) of the data X."""
        sample_name = self.preprocessor.sample_name
        X0 = X.copy(deep=True)
        Xtau = X.shift({sample_name: -tau}).dropna(sample_name)

        X0 = X0.rename({"mode": "feature1"})
        Xtau = Xtau.rename({"mode": "feature2"})

        n_samples = Xtau[sample_name].size
        return xr.dot(X0, Xtau, dims=[sample_name]) / (n_samples - 1)

    @staticmethod
    def _compute_matrix_inverse(X, dims):
        """Compute the inverse of a symmetric matrix X."""
        return xr.apply_ufunc(
            np.linalg.inv,
            X,
            input_core_dims=[dims],
            output_core_dims=[dims[::-1]],
            vectorize=False,
            dask="allowed",
        )

    def _fit_algorithm(self, X: DataArray) -> Self:
        assert_not_complex(X)

        sample_name = self.sample_name
        feature_name = self.feature_name

        # Perform PCA as a pre-processing step
        pca = EOF(
            n_modes=self._params["n_pca_modes"],
            standardize=False,
            use_coslat=False,
            sample_name=self.sample_name,
            feature_name=self.feature_name,
            solver=self._params["solver"],
            compute=self._params["compute"],
            random_state=self._params["random_state"],
            check_nans=False,
            solver_kwargs=self._params["solver_kwargs"],
        )
        pca.fit(X, dim=sample_name)
        n_samples = X.coords[sample_name].size
        comps = pca.data["components"] * np.sqrt(n_samples - 1)
        # -> comps (feature x mode)
        scores = pca.data["scores"] / np.sqrt(n_samples - 1)
        # -> scores (sample x mode)

        # Compute the covariance matrix with zero time lag
        C0 = self._Ctau(scores, 0)
        # -> C0 (feature1 x feature2)
        # C0inv = self._compute_matrix_inverse(C0, dims=("feature1", "feature2"))
        # -> C0inv (feature2 x feature1)
        M = 0.5 * C0
        # -> M (feature1 x feature2)
        tau_max = self._params["tau_max"]
        for tau in range(1, tau_max + 1):
            Ctau = self._Ctau(scores, tau)
            if tau == tau_max:
                Ctau = 0.5 * Ctau
            M = M + (Ctau)

        MT = xr.DataArray(M.data.T, dims=M.dims, coords=M.coords)
        # -> MT (feature1 x feature2)
        M_summed = M + MT
        # -> M_summed (feature1 x feature2)

        # Instead of solving the generalized eigenvalue problem
        # as proposed in DelSole (2001), we solve the
        # eigenvalue problem of the alternativ formulation
        # using a symmtric matrix given in
        # A. Hannachi (2021), Patterns Identification and
        # Data Mining in Weather and Climate, Equation (8.20)
        decomposer = Decomposer(
            n_modes=C0.shape[0],
            flip_signs=False,
            compute=self._params["compute"],
            solver="full",
        )
        decomposer.fit(C0, dims=("feature1", "feature2"))
        C0_sqrt = decomposer.U_ * np.sqrt(decomposer.s_)
        # -> C0_sqrt (feature1 x mode)
        C0_sqrt_inv = self._compute_matrix_inverse(C0_sqrt, dims=("feature1", "mode"))
        # -> C0_sqrt_inv (mode x feature1)
        target = 0.5 * xr.dot(C0_sqrt_inv, M_summed, dims="feature1")
        # -> target (mode x feature2)
        target = xr.dot(
            target, C0_sqrt_inv.rename({"mode": "feature2"}), dims="feature2"
        )
        # -> target (mode x feature1)
        target = target.rename({"feature1": "dummy"})
        target = target.rename({"mode": "feature1"})
        # -> target (feature1 x dummy)

        # Solve the symmetric eigenvalue problem
        eigensolver = Decomposer(
            n_modes=self._params["n_modes"], flip_signs=False, solver="full"
        )
        eigensolver.fit(target, dims=("feature1", "dummy"))
        U = eigensolver.U_
        # -> U (feature1 x mode)
        lbda = eigensolver.s_
        # -> lbda (mode)
        # U, lbda, ct = xr.apply_ufunc(
        #     np.linalg.svd,
        #     target,
        #     input_core_dims=[("feature1", "dummy")],
        #     output_core_dims=[("feature1", "mode"), ("mode",), ("mode", "dummy")],
        #     vectorize=False,
        #     dask="allowed",
        # )
        # Compute the filter patterns
        V = C0_sqrt_inv.rename({"mode": "mode1"}).dot(
            U.rename({"mode": "mode2"}), dims="feature1"
        )
        # -> V (mode1 x mode2)

        # Compute the optimally persistent patterns (OPPs)
        W = xr.dot(
            C0.rename({"feature2": "temp"}), V.rename({"mode1": "temp"}), dims="temp"
        )
        # -> W (feature1 x mode2)

        # Compute the time series of the optimally persistent patterns (OPPs)
        P = xr.dot(scores.rename({"mode": "mode1"}), V, dims="mode1")
        # -> P (sample x mode2)

        # Transform filter patterns and OPPs into original space
        V = xr.dot(comps.rename({"mode": "mode1"}), V, dims="mode1")
        # -> V (feature x mode2)

        W = xr.dot(comps.rename({"mode": "feature1"}), W, dims="feature1")
        # -> W (feature x mode2)

        # Rename dimensions
        U = U.rename({"feature1": feature_name})  # -> (feature x mode)
        V = V.rename({"mode2": "mode"})  # -> (feature x mode)
        W = W.rename({"mode2": "mode"})  # -> (feature x mode)
        P = P.rename({"mode2": "mode"})  # -> (sample x mode)
        scores = scores.rename({"mode": feature_name})  # -> (sample x feature)

        # Compute the norms of the scores
        norms = xr.apply_ufunc(
            np.linalg.norm,
            P,
            input_core_dims=[["sample"]],
            vectorize=False,
            dask="allowed",
            kwargs={"axis": -1},
        )

        # Store the results
        # NOTE: not sure if "scores" should be taken as input data here, "data" may be more correct -> to be verified
        self.data.add(name="input_data", data=scores, allow_compute=False)
        self.data.add(name="components", data=W, allow_compute=True)
        self.data.add(name="scores", data=P, allow_compute=True)
        self.data.add(name="norms", data=norms, allow_compute=True)
        self.data.add(name="filter_patterns", data=V, allow_compute=True)
        self.data.add(name="decorrelation_time", data=lbda, allow_compute=True)

        self.data.set_attrs(self.attrs)
        self._U = U  # store U for testing purposes of orthogonality
        self._C0 = C0  # store C0 for testing purposes of orthogonality
        return self

    def _transform_algorithm(self, X: DataArray) -> DataArray:
        raise NotImplementedError("OPA does not (yet) support transform()")

    def _inverse_transform_algorithm(self, scores) -> DataObject:
        raise NotImplementedError("OPA does not (yet) support inverse_transform()")

    def components(self) -> DataObject:
        """Return the optimally persistent patterns (OPPs)."""
        return super().components()

    def scores(self) -> DataArray:
        """Return the time series of the OPPs.

        The time series have a maximum decorrelation time that are uncorrelated with each other.
        """
        return super().scores()

    def decorrelation_time(self) -> DataArray:
        """Return the decorrelation time of the optimal persistence pattern (OPP)."""
        return self.data["decorrelation_time"]

    def filter_patterns(self) -> DataObject:
        """Return the filter patterns."""
        fps = self.data["filter_patterns"]
        return self.preprocessor.inverse_transform_components(fps)



import warnings

import numpy as np
import xarray as xr
from typing_extensions import Self

from ..linalg._numpy import _fractional_matrix_power
from ..utils.data_types import (
    DataArray,
    Dims,
    DimsList,
)
from ..utils.sanity_checks import assert_single_dataarray
from .transformer import Transformer


class Whitener(Transformer):
    """Fractional whitening of 2D DataArray.

    For ``alpha=1`` (no whiteneing), it just becomes the identity transformation.

    Parameters
    ----------
    alpha: float, default=0
        Power parameter to perform fractional whitening, where 0 corresponds to full whitening (standardized and decorrelated) and 1 to no whitening.
    sample_name: str, default="sample"
        Name of the sample dimension.
    feature_name: str, default="feature"
        Name of the feature dimension.
    random_state: int | None, default=None
        Random seed for reproducibility.
    """

    def __init__(
        self,
        alpha: float = 0,
        sample_name: str = "sample",
        feature_name: str = "feature",
        random_state: int | None = None,
    ):
        super().__init__(sample_name, feature_name)

        # Verify that alpha has a lower bound of 0
        if alpha < 0:
            raise ValueError("`alpha` must be greater than or equal to 0")

        alpha = float(alpha)

        self.alpha = alpha
        self.random_state = random_state

        # Check whether Whitener is identity transformation
        self.is_identity = self._check_identity_transform()

    def _check_identity_transform(self) -> bool:
        eps = np.finfo(self.alpha).eps
        alpha_is_one = (1.0 - self.alpha) < eps
        if alpha_is_one:
            return True
        else:
            return False

    def _sanity_check_input(self, X) -> None:
        assert_single_dataarray(X)

        if len(X.dims) != 2:
            raise ValueError("Input DataArray must have shape 2")

        if X.dims != (self.sample_name, self.feature_name):
            raise ValueError(
                "Input DataArray must have dimensions ({:}, {:})".format(
                    self.sample_name, self.feature_name
                )
            )

    def get_serialization_attrs(self) -> dict:
        return dict(
            alpha=self.alpha,
            is_identity=self.is_identity,
            T=self.T,
            Tinv=self.Tinv,
            feature_name=self.feature_name,
            n_samples=self.n_samples,
        )

    def fit(
        self,
        X: DataArray,
        sample_dims: Dims | None = None,
        feature_dims: DimsList | None = None,
    ) -> Self:
        self._sanity_check_input(X)
        n_samples, n_features = X.shape
        self.n_samples = n_samples

        if self.is_identity:
            self.T = xr.DataArray(1, name="identity")
            self.Tinv = xr.DataArray(1, name="identity")

        # Compute the fractional whitening transformation directly based on covariance matrix
        else:
            if n_samples < n_features:
                warnings.warn(
                    f"The number of samples ({n_samples}) is smaller than the number of features ({n_features}), leading to an ill-conditioned problem. This may cause unstable results. Consider using PCA to reduce dimensionality and stabilize the problem by setting `use_pca=True`."
                )

            self.T, self.Tinv = self._compute_whitener_transform(X)

        return self

    def _compute_whitener_transform(self, X: DataArray) -> tuple[DataArray, DataArray]:
        T, Tinv = xr.apply_ufunc(
            self._compute_whitener_transform_numpy,
            X,
            input_core_dims=[[self.sample_name, self.feature_name]],
            output_core_dims=[[self.feature_name, "mode"], ["mode", self.feature_name]],
            dask="allowed",
        )
        T = T.assign_coords(mode=X.coords[self.feature_name].data)
        Tinv = Tinv.assign_coords(mode=X.coords[self.feature_name].data)
        return T, Tinv

    def _compute_whitener_transform_numpy(self, X):
        nc = X.shape[0]
        C = X.conj().T @ X / nc
        power = (self.alpha - 1) / 2
        svd_kwargs = {"random_state": self.random_state, "solver": "full"}
        T = _fractional_matrix_power(C, power, **svd_kwargs)
        try:
            Tinv = np.linalg.inv(T)
        except np.linalg.LinAlgError:
            Tinv = np.linalg.pinv(T)
        return T, Tinv

    def transform(self, X: DataArray) -> DataArray:
        """Transform new data into the fractional whitened space."""

        self._sanity_check_input(X)
        if self.is_identity:
            return X
        else:
            transformed = xr.dot(X, self.T, dims=self.feature_name)
            transformed.name = X.name
            return transformed.rename({"mode": self.feature_name})

    def fit_transform(
        self,
        X: DataArray,
        sample_dims: Dims | None = None,
        feature_dims: DimsList | None = None,
    ) -> DataArray:
        return self.fit(X, sample_dims, feature_dims).transform(X)

    def inverse_transform_data(self, X: DataArray) -> DataArray:
        """Transform 2D data (sample x feature) from whitened space back into original space.

        Parameters
        ----------
        X: DataArray
            Data to transform back into original space.
        """
        if self.is_identity:
            return X
        else:
            X = X.rename({self.feature_name: "mode"})
            return xr.dot(X, self.Tinv, dims="mode")

    def transform_components(self, X: DataArray) -> DataArray:
        """Transform 2D components (feature x mode) into whitened space."""

        if self.is_identity:
            return X
        else:
            dummy_dim = "dummy_dim"
            VS = self.T.conj().T
            VS = VS.rename({"mode": dummy_dim})
            transformed = xr.dot(VS, X, dims=self.feature_name)
            return transformed.rename({dummy_dim: self.feature_name})

    def inverse_transform_components(self, X: DataArray) -> DataArray:
        """Transform 2D components (feature x mode) from whitened space back into original space."""

        if self.is_identity:
            return X
        else:
            dummy_dim = "dummy_dim"
            comps_pc_space = X.rename({self.feature_name: dummy_dim})
            VS = self.Tinv.conj().T
            VS = VS.rename({"mode": dummy_dim})
            return xr.dot(VS, comps_pc_space, dims=dummy_dim)

    def inverse_transform_scores(self, X: DataArray) -> DataArray:
        """Transform 2D scores (sample x mode) from whitened space back into original space."""

        return X

    def inverse_transform_scores_unseen(self, X: DataArray) -> DataArray:
        """Transform unseen 2D scores (sample x mode) from whitened space back into original space."""

        return X



from typing import Any, Generic, Type, TypeVar

from typing_extensions import Self

from ..utils.data_types import (
    Data,
    DataArray,
    DataVar,
    Dims,
    DimsList,
)
from .dimension_renamer import DimensionRenamer
from .multi_index_converter import MultiIndexConverter
from .sanitizer import Sanitizer
from .scaler import Scaler
from .stacker import Stacker

T = TypeVar(
    "T",
    bound=(DimensionRenamer | Scaler | MultiIndexConverter | Stacker | Sanitizer),
)


class GenericListTransformer(Generic[T]):
    """Apply a Transformer to each of the elements of a list.

    Parameters
    ----------
    transformer: Transformer
        Transformer class to apply to list elements.
    kwargs: dict
        Keyword arguments for the transformer.
    """

    def __init__(self, transformer: Type[T], **kwargs):
        self.transformer_class = transformer
        self.transformers: list[T] = []
        self.init_kwargs = kwargs

    def fit(
        self,
        X: list[DataVar],
        sample_dims: Dims,
        feature_dims: DimsList,
        iter_kwargs: dict[str, list[Any]] = {},
    ) -> Self:
        """Fit transformer to each data element in the list.

        Parameters
        ----------
        X: list[Data]
            list of data elements.
        sample_dims: Dims
            Sample dimensions.
        feature_dims: DimsList
            Feature dimensions.
        iter_kwargs: dict[str, list[Any]]
            Keyword arguments for the transformer that should be iterated over.

        """
        self._sample_dims = sample_dims
        self._feature_dims = feature_dims
        self._iter_kwargs = iter_kwargs

        for i, x in enumerate(X):
            # Add transformer specific keyword arguments
            # For iterable kwargs, use the i-th element of the iterable
            kwargs = {k: v[i] for k, v in self._iter_kwargs.items()}
            proc: T = self.transformer_class(**self.init_kwargs)
            proc.fit(x, sample_dims, feature_dims[i], **kwargs)
            self.transformers.append(proc)
        return self

    def transform(self, X: list[Data]) -> list[Data]:
        X_transformed: list[Data] = []
        for x, proc in zip(X, self.transformers):
            X_transformed.append(proc.transform(x))  #  type: ignore
        return X_transformed

    def fit_transform(
        self,
        X: list[Data],
        sample_dims: Dims,
        feature_dims: DimsList,
        iter_kwargs: dict[str, list[Any]] = {},
    ) -> list[Data]:
        return self.fit(X, sample_dims, feature_dims, iter_kwargs).transform(X)  # type: ignore

    def inverse_transform_data(self, X: list[Data]) -> list[Data]:
        X_inverse_transformed: list[Data] = []
        for x, proc in zip(X, self.transformers):
            x_inv_trans = proc.inverse_transform_data(x)  #  type: ignore
            X_inverse_transformed.append(x_inv_trans)
        return X_inverse_transformed

    def inverse_transform_components(self, X: list[Data]) -> list[Data]:
        X_inverse_transformed: list[Data] = []
        for x, proc in zip(X, self.transformers):
            x_inv_trans = proc.inverse_transform_components(x)  #  type: ignore
            X_inverse_transformed.append(x_inv_trans)
        return X_inverse_transformed

    def inverse_transform_scores(self, X: DataArray) -> DataArray:
        return self.transformers[0].inverse_transform_scores(X)

    def inverse_transform_scores_unseen(self, X: DataArray) -> DataArray:
        return self.transformers[0].inverse_transform_scores_unseen(X)



import dask.array as da
import numpy as np
import pytest
import xarray as xr

from xeofs.cross import RDA

from ...utilities import skip_if_missing_engine


def generate_random_data(shape, lazy=False, seed=142):
    rng = np.random.default_rng(seed)
    if lazy:
        return xr.DataArray(
            da.random.random(shape, chunks=(5, 5)),
            dims=["sample", "feature"],
            coords={"sample": np.arange(shape[0]), "feature": np.arange(shape[1])},
        )
    else:
        return xr.DataArray(
            rng.random(shape),
            dims=["sample", "feature"],
            coords={"sample": np.arange(shape[0]), "feature": np.arange(shape[1])},
        )


def generate_well_conditioned_data(lazy=False):
    rng = np.random.default_rng(142)
    t = np.linspace(0, 50, 200)
    std = 0.1
    x1 = np.sin(t)[:, None] + rng.normal(0, std, size=(200, 2))
    x2 = np.sin(t)[:, None] + rng.normal(0, std, size=(200, 3))
    x1[:, 1] = x1[:, 1] ** 2
    x2[:, 1] = x2[:, 1] ** 3
    x2[:, 2] = abs(x2[:, 2]) ** (0.5)
    coords_time = np.arange(len(t))
    coords_fx = [1, 2]
    coords_fy = [1, 2, 3]
    X = xr.DataArray(
        x1,
        dims=["sample", "feature"],
        coords={"sample": coords_time, "feature": coords_fx},
    )
    Y = xr.DataArray(
        x2,
        dims=["sample", "feature"],
        coords={"sample": coords_time, "feature": coords_fy},
    )
    if lazy:
        X = X.chunk({"sample": 5, "feature": -1})
        Y = Y.chunk({"sample": 5, "feature": -1})
        return X, Y
    else:
        return X, Y


@pytest.fixture
def rda():
    return RDA(n_modes=1)


def test_initialization():
    model = RDA()
    assert model is not None


def test_fit(rda):
    X, Y = generate_well_conditioned_data()
    rda.fit(X, Y, dim="sample")
    assert hasattr(rda, "preprocessor1")
    assert hasattr(rda, "preprocessor2")
    assert hasattr(rda, "data")


def test_fit_empty_data(rda):
    with pytest.raises(ValueError):
        rda.fit(xr.DataArray(), xr.DataArray(), "time")


def test_fit_invalid_dims(rda):
    X, Y = generate_well_conditioned_data()
    with pytest.raises(ValueError):
        rda.fit(X, Y, dim=("invalid_dim1", "invalid_dim2"))


def test_transform(rda):
    X, Y = generate_well_conditioned_data()
    rda.fit(X, Y, dim="sample")
    result = rda.transform(X, Y)
    assert isinstance(result, list)
    assert isinstance(result[0], xr.DataArray)


def test_transform_unseen_data(rda):
    X, Y = generate_well_conditioned_data()
    x = X.isel(sample=slice(151, 200))
    y = Y.isel(sample=slice(151, 200))
    X = X.isel(sample=slice(None, 150))
    Y = Y.isel(sample=slice(None, 150))

    rda.fit(X, Y, "sample")
    result = rda.transform(x, y)
    assert isinstance(result, list)
    assert isinstance(result[0], xr.DataArray)
    # Check that unseen data can be transformed
    assert result[0].notnull().all()
    assert result[1].notnull().all()


def test_inverse_transform(rda):
    X, Y = generate_well_conditioned_data()
    rda.fit(X, Y, "sample")
    # Assuming mode as 1 for simplicity
    scores1 = rda.data["scores1"].sel(mode=1)
    scores2 = rda.data["scores2"].sel(mode=1)
    Xrec1, Xrec2 = rda.inverse_transform(scores1, scores2)
    assert isinstance(Xrec1, xr.DataArray)
    assert isinstance(Xrec2, xr.DataArray)


@pytest.mark.parametrize("use_pca", [False, True])
def test_squared_covariance_fraction(use_pca):
    X, Y = generate_well_conditioned_data()
    rda = RDA(n_modes=2, use_pca=use_pca, n_pca_modes="all")
    rda.fit(X, Y, "sample")
    scf = rda.squared_covariance_fraction()
    assert isinstance(scf, xr.DataArray)
    assert all(scf <= 1), "Squared covariance fraction is greater than 1"


@pytest.mark.parametrize("use_pca", [False, True])
def test_total_squared_covariance(use_pca):
    X, Y = generate_well_conditioned_data()

    # Compute total squared covariance
    X_ = X.rename({"feature": "x"})
    Y_ = Y.rename({"feature": "y"})
    cov_mat = xr.cov(X_, Y_, dim="sample")
    tsc = (cov_mat**2).sum()

    rda = RDA(n_modes=2, use_pca=use_pca, n_pca_modes="all")
    rda.fit(X, Y, "sample")
    tsc_model = rda.data["total_squared_covariance"]
    xr.testing.assert_allclose(tsc, tsc_model)


def test_fit_different_coordinates():
    """Like a lagged RDA scenario"""
    X, Y = generate_well_conditioned_data()
    X = X.isel(sample=slice(0, 99))
    Y = Y.isel(sample=slice(100, 199))
    rda = RDA(n_modes=2, use_pca=False)
    rda.fit(X, Y, "sample")
    r = rda.cross_correlation_coefficients()
    # Correlation coefficents are not zero
    assert np.all(r > np.finfo(r.dtype).eps)


@pytest.mark.parametrize(
    "dim",
    [(("time",)), (("lat", "lon")), (("lon", "lat"))],
)
def test_components(mock_data_array, dim):
    rda = RDA(n_modes=2, use_pca=False)
    rda.fit(mock_data_array, mock_data_array, dim)
    components1, components2 = rda.components()
    feature_dims = tuple(set(mock_data_array.dims) - set(dim))
    assert isinstance(components1, xr.DataArray)
    assert isinstance(components2, xr.DataArray)
    assert set(components1.dims) == set(
        ("mode",) + feature_dims
    ), "Components1 does not have the right feature dimensions"
    assert set(components2.dims) == set(
        ("mode",) + feature_dims
    ), "Components2 does not have the right feature dimensions"


@pytest.mark.parametrize("shapeX", [(30, 10)])
@pytest.mark.parametrize("shapeY", [(30, 10), (30, 5), (30, 15)])
@pytest.mark.parametrize("use_pca", [False, True])
def test_components_coordinates(shapeX, shapeY, use_pca):
    # Test that the components have the right coordinates
    X = generate_random_data(shapeX)
    Y = generate_random_data(shapeY)

    rda = RDA(n_modes=2, use_pca=use_pca, n_pca_modes="all")
    rda.fit(X, Y, "sample")
    components1, components2 = rda.components()
    xr.testing.assert_equal(components1.coords["feature"], X.coords["feature"])
    xr.testing.assert_equal(components2.coords["feature"], Y.coords["feature"])


@pytest.mark.parametrize("correction", [(None), ("fdr_bh")])
def test_homogeneous_patterns(correction):
    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    rda = RDA(n_modes=10, use_pca=False)
    rda.fit(X, Y, "sample")

    _ = rda.homogeneous_patterns(correction=correction)


@pytest.mark.parametrize(
    "correction",
    [(None), ("fdr_bh")],
)
def test_heterogeneous_patterns(correction):
    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    rda = RDA(n_modes=10, use_pca=False)
    rda.fit(X, Y, "sample")

    _ = rda.heterogeneous_patterns(correction=correction)


def test_predict():
    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    rda = RDA(n_modes=10, use_pca=False)
    rda.fit(X, Y, "sample")

    Xnew = generate_random_data((200, 10), seed=123)

    Ry_pred = rda.predict(Xnew)
    _ = rda.inverse_transform(Y=Ry_pred)


@pytest.mark.parametrize("engine", ["h5netcdf", "netcdf4", "zarr"])
def test_save_load(tmp_path, engine):
    """Test save/load methods in MCA class, ensuring that we can
    roundtrip the model and get the same results when transforming
    data."""
    skip_if_missing_engine(engine)

    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    original = RDA()
    original.fit(X, Y, "sample")

    # Save the RDA model
    original.save(tmp_path / "rda", engine=engine)

    # Check that the RDA model has been saved
    assert (tmp_path / "rda").exists()

    # Recreate the model from saved file
    loaded = RDA.load(tmp_path / "rda", engine=engine)

    # Check that the params and DataContainer objects match
    assert original.get_params() == loaded.get_params()
    assert all([key in loaded.data for key in original.data])
    for key in original.data:
        if original.data._allow_compute[key]:
            assert loaded.data[key].equals(original.data[key])
        else:
            # but ensure that input data is not saved by default
            assert loaded.data[key].size <= 1
            assert loaded.data[key].attrs["placeholder"] is True

    # Test that the recreated model can be used to transform new data
    assert np.allclose(
        original.transform(X, Y),
        loaded.transform(X, Y),
    )

    # The loaded model should also be able to inverse_transform new data
    XYr_o = original.inverse_transform(*original.scores())
    XYr_l = loaded.inverse_transform(*loaded.scores())
    assert np.allclose(XYr_o[0], XYr_l[0])
    assert np.allclose(XYr_o[1], XYr_l[1])


def test_serialize_deserialize_dataarray(mock_data_array):
    """Test roundtrip serialization when the model is fit on a DataArray."""
    model = RDA()
    model.fit(mock_data_array, mock_data_array, "time")
    dt = model.serialize()
    rebuilt_model = RDA.deserialize(dt)
    assert np.allclose(
        model.transform(mock_data_array), rebuilt_model.transform(mock_data_array)
    )


def test_serialize_deserialize_dataset(mock_dataset):
    """Test roundtrip serialization when the model is fit on a Dataset."""
    model = RDA()
    model.fit(mock_dataset, mock_dataset, "time")
    dt = model.serialize()
    rebuilt_model = RDA.deserialize(dt)
    assert np.allclose(
        model.transform(mock_dataset), rebuilt_model.transform(mock_dataset)
    )



import pytest

from xeofs.preprocessing.dimension_renamer import DimensionRenamer
from ..utilities import (
    data_is_dask,
    get_dims_from_data,
)

# =============================================================================
# GENERALLY VALID TEST CASES
# =============================================================================
N_VARIABLES = [1, 2]
N_SAMPLE_DIMS = [1, 2]
N_FEATURE_DIMS = [1, 2]
INDEX_POLICY = ["index", "multiindex"]
NAN_POLICY = ["no_nan", "fulldim"]
DASK_POLICY = ["no_dask", "dask"]
SEED = [0]

VALID_TEST_DATA = [
    (nv, ns, nf, index, nan, dask)
    for nv in N_VARIABLES
    for ns in N_SAMPLE_DIMS
    for nf in N_FEATURE_DIMS
    for index in INDEX_POLICY
    for nan in NAN_POLICY
    for dask in DASK_POLICY
]


# TESTS
# =============================================================================
@pytest.mark.parametrize(
    "synthetic_dataset",
    VALID_TEST_DATA,
    indirect=["synthetic_dataset"],
)
def test_transform(synthetic_dataset):
    all_dims, sample_dims, feature_dims = get_dims_from_data(synthetic_dataset)

    n_dims = len(all_dims)

    base = "new"
    start = 10
    expected_dims = set(base + str(i) for i in range(start, start + n_dims))

    renamer = DimensionRenamer(base=base, start=start)
    renamer.fit(synthetic_dataset, sample_dims, feature_dims)
    transformed_data = renamer.transform(synthetic_dataset)

    is_dask_before = data_is_dask(synthetic_dataset)
    is_dask_after = data_is_dask(transformed_data)

    # Transforming doesn't change the dask-ness of the data
    assert is_dask_before == is_dask_after

    # Transforming converts dimension names
    given_dims = set(transformed_data.dims)
    assert given_dims == expected_dims

    # Result is robust to calling the method multiple times
    transformed_data = renamer.transform(synthetic_dataset)
    given_dims = set(transformed_data.dims)
    assert given_dims == expected_dims


@pytest.mark.parametrize(
    "synthetic_dataset",
    VALID_TEST_DATA,
    indirect=["synthetic_dataset"],
)
def test_inverse_transform_data(synthetic_dataset):
    all_dims, sample_dims, feature_dims = get_dims_from_data(synthetic_dataset)

    base = "new"
    start = 10

    renamer = DimensionRenamer(base=base, start=start)
    renamer.fit(synthetic_dataset, sample_dims, feature_dims)
    transformed_data = renamer.transform(synthetic_dataset)
    inverse_transformed_data = renamer.inverse_transform_data(transformed_data)

    is_dask_before = data_is_dask(synthetic_dataset)
    is_dask_after = data_is_dask(transformed_data)

    # Transforming doesn't change the dask-ness of the data
    assert is_dask_before == is_dask_after

    assert inverse_transformed_data.identical(synthetic_dataset)
    assert set(inverse_transformed_data.dims) == set(synthetic_dataset.dims)



import numpy as np
import pytest
import xarray as xr

from xeofs.single import ExtendedEOF


def test_init():
    """Tests the initialization of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)

    # Assert preprocessor has been initialized
    assert hasattr(eof, "_params")
    assert hasattr(eof, "preprocessor")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit(dim, mock_data_array):
    """Tests the fit method of the ExtendedEOF class"""

    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array, dim)

    # Assert the required attributes have been set
    assert hasattr(eof, "preprocessor")
    assert hasattr(eof, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_singular_values(dim, mock_data_array):
    """Tests the singular_values method of the ExtendedEOF class"""

    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array, dim)

    # Test singular_values method
    singular_values = eof.singular_values()
    assert isinstance(singular_values, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_explained_variance(dim, mock_data_array):
    """Tests the explained_variance method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array, dim)

    # Test explained_variance method
    explained_variance = eof.explained_variance()
    assert isinstance(explained_variance, xr.DataArray)
    # Explained variance must be positive
    assert (explained_variance > 0).all()


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_explained_variance_ratio(dim, mock_data_array):
    """Tests the explained_variance_ratio method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array, dim)

    # Test explained_variance_ratio method
    explained_variance_ratio = eof.explained_variance_ratio()
    assert isinstance(explained_variance_ratio, xr.DataArray)
    # Explained variance ratio must be positive
    assert (
        explained_variance_ratio > 0
    ).all(), "The explained variance ratio must be positive"
    # The sum of the explained variance ratio must be <= 1
    assert (
        explained_variance_ratio.sum() <= 1 + 1e-5
    ), "The sum of the explained variance ratio must be <= 1"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_isolated_nans(dim, mock_data_array_isolated_nans):
    """Tests the components method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    with pytest.raises(ValueError):
        eof.fit(mock_data_array_isolated_nans, dim)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components(dim, mock_data_array):
    """Tests the components method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array, dim)

    # Test components method
    components = eof.components()
    feature_dims = tuple(set(mock_data_array.dims) - set(dim))
    assert isinstance(components, xr.DataArray), "Components is not a DataArray"
    given_dims = set(components.dims)
    expected_dims = set(feature_dims + ("mode", "embedding"))
    assert (
        given_dims == expected_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_fulldim_nans(dim, mock_data_array_full_dimensional_nans):
    """Tests the components method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array_full_dimensional_nans, dim)

    # Test components method
    components = eof.components()
    feature_dims = tuple(set(mock_data_array_full_dimensional_nans.dims) - set(dim))
    assert isinstance(components, xr.DataArray), "Components is not a DataArray"
    given_dims = set(components.dims)
    expected_dims = set(feature_dims + ("mode", "embedding"))
    assert (
        given_dims == expected_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_boundary_nans(dim, mock_data_array_boundary_nans):
    """Tests the components method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array_boundary_nans, dim)

    # Test components method
    components = eof.components()
    feature_dims = tuple(set(mock_data_array_boundary_nans.dims) - set(dim))
    assert isinstance(components, xr.DataArray), "Components is not a DataArray"
    given_dims = set(components.dims)
    expected_dims = set(feature_dims + ("mode", "embedding"))
    assert (
        given_dims == expected_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_dataset(dim, mock_dataset):
    """Tests the components method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_dataset, dim)

    # Test components method
    components = eof.components()
    feature_dims = tuple(set(mock_dataset.dims) - set(dim))
    assert isinstance(components, xr.Dataset), "Components is not a Dataset"
    assert set(components.data_vars) == set(
        mock_dataset.data_vars
    ), "Components does not have the same data variables as the input Dataset"
    given_dims = set(components.dims)
    expected_dims = set(feature_dims + ("mode", "embedding"))
    assert (
        given_dims == expected_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_dataarray_list(dim, mock_data_array_list):
    """Tests the components method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array_list, dim)

    # Test components method
    components = eof.components()
    feature_dims = [tuple(set(data.dims) - set(dim)) for data in mock_data_array_list]
    assert isinstance(components, list), "Components is not a list"
    assert len(components) == len(
        mock_data_array_list
    ), "Components does not have the same length as the input list"
    assert isinstance(
        components[0], xr.DataArray
    ), "Components is not a list of DataArrays"
    for comp, feat_dims in zip(components, feature_dims):
        given_dims = set(comp.dims)
        expected_dims = set(feat_dims + ("mode", "embedding"))
        assert (
            given_dims == expected_dims
        ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores(dim, mock_data_array):
    """Tests the scores method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array, dim)

    # Test scores method
    scores = eof.scores()
    assert isinstance(scores, xr.DataArray), "Scores is not a DataArray"
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_fulldim_nans(dim, mock_data_array_full_dimensional_nans):
    """Tests the scores method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array_full_dimensional_nans, dim)

    # Test scores method
    scores = eof.scores()
    assert isinstance(scores, xr.DataArray), "Scores is not a DataArray"
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_boundary_nans(dim, mock_data_array_boundary_nans):
    """Tests the scores method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array_boundary_nans, dim)

    # Test scores method
    scores = eof.scores()
    assert isinstance(scores, xr.DataArray), "Scores is not a DataArray"
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_dataset(dim, mock_dataset):
    """Tests the scores method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_dataset, dim)

    # Test scores method
    scores = eof.scores()
    assert isinstance(scores, xr.DataArray)
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_dataarray_list(dim, mock_data_array_list):
    """Tests the scores method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)
    eof.fit(mock_data_array_list, dim)

    # Test scores method
    scores = eof.scores()
    assert isinstance(scores, xr.DataArray)
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


def test_get_params():
    """Tests the get_params method of the ExtendedEOF class"""
    eof = ExtendedEOF(n_modes=5, tau=2, embedding=2)

    # Test get_params method
    params = eof.get_params()
    assert isinstance(params, dict)
    assert params.get("n_modes") == 5
    assert params.get("tau") == 2
    assert params.get("embedding") == 2
    assert params.get("solver") == "auto"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_transform(dim, mock_data_array):
    """Test projecting new unseen data onto the components (EOFs/eigenvectors)"""

    # Create a xarray DataArray with random data
    model = ExtendedEOF(n_modes=5, tau=2, embedding=2, solver="full")
    model.fit(mock_data_array, dim)
    model.scores()

    # Create a new xarray DataArray with random data
    new_data = mock_data_array

    with pytest.raises(NotImplementedError):
        model.transform(new_data)

    # # Check that the projection has the right dimensions
    # assert projections.dims == scores.dims, "Projection has wrong dimensions"  # type: ignore

    # # Check that the projection has the right data type
    # assert isinstance(projections, xr.DataArray), "Projection is not a DataArray"

    # # Check that the projection has the right name
    # assert projections.name == "scores", "Projection has wrong name: {}".format(
    #     projections.name
    # )

    # # Check that the projection's data is the same as the scores
    # np.testing.assert_allclose(
    #     scores.sel(mode=slice(1, 3)), projections.sel(mode=slice(1, 3)), rtol=1e-3
    # )


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_inverse_transform(dim, mock_data_array):
    """Test inverse_transform method in ExtendedEOF class."""

    # instantiate the ExtendedEOF class with necessary parameters
    eeof = ExtendedEOF(n_modes=5, tau=2, embedding=2)

    # fit the ExtendedEOF model
    eeof.fit(mock_data_array, dim=dim)
    scores = eeof.scores()

    # Test with scalar
    mode = 1
    reconstructed_data = eeof.inverse_transform(scores.sel(mode=mode))
    assert isinstance(reconstructed_data, xr.DataArray)

    # Test with slice
    mode = slice(1, 2)
    reconstructed_data = eeof.inverse_transform(scores.sel(mode=mode))
    assert isinstance(reconstructed_data, xr.DataArray)

    # Test with array of tick labels
    mode = np.array([1, 3])
    reconstructed_data = eeof.inverse_transform(scores.sel(mode=mode))
    assert isinstance(reconstructed_data, xr.DataArray)

    # Check that the reconstructed data has the same dimensions as the original data
    assert set(reconstructed_data.dims) == set(mock_data_array.dims)



import warnings

import numpy as np
import pytest

from xeofs.single import HilbertEOF

warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit(mock_data_array, dim):
    """Test fitting a HilbertEOF model"""
    # Create a xarray DataArray with random data
    ceof = HilbertEOF(n_modes=2)
    ceof.fit(mock_data_array, dim)

    # Check that the fit method has properly populated the attributes
    assert hasattr(ceof, "preprocessor")
    assert hasattr(ceof, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_amplitude(mock_data_array, dim):
    """Test computation of components amplitude in HilbertEOF model"""
    ceof = HilbertEOF(n_modes=2)
    ceof.fit(mock_data_array, dim)

    comp_amp = ceof.components_amplitude()
    assert comp_amp is not None
    assert (comp_amp.fillna(0) >= 0).all()  # type: ignore


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_phase(mock_data_array, dim):
    """Test computation of components phase in HilbertEOF model"""
    ceof = HilbertEOF(n_modes=2)
    ceof.fit(mock_data_array, dim)

    comp_phase = ceof.components_phase()
    assert comp_phase is not None
    assert ((-np.pi <= comp_phase.fillna(0)) & (comp_phase.fillna(0) <= np.pi)).all()  # type: ignore


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_amplitude(mock_data_array, dim):
    """Test computation of scores amplitude in HilbertEOF model"""
    ceof = HilbertEOF(n_modes=2)
    ceof.fit(mock_data_array, dim)

    scores_amp = ceof.scores_amplitude()
    assert scores_amp is not None
    assert (scores_amp.fillna(0) >= 0).all()  # type: ignore


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_phase(mock_data_array, dim):
    """Test computation of scores phase in HilbertEOF model"""
    ceof = HilbertEOF(n_modes=2)
    ceof.fit(mock_data_array, dim)

    scores_phase = ceof.scores_phase()
    assert scores_phase is not None
    assert (
        (-np.pi <= scores_phase.fillna(0)) & (scores_phase.fillna(0) <= np.pi)
    ).all()  # type: ignore


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_compute(mock_dask_data_array, dim):
    """Test computation of all attributes in HilbertEOF model"""
    ceof = HilbertEOF(n_modes=2)
    with pytest.raises(NotImplementedError):
        ceof.fit(mock_dask_data_array, dim)



import numpy as np
import pytest
import xarray as xr

# Import the classes from your modules
from xeofs.cross import HilbertMCA, HilbertMCARotator

from ...utilities import skip_if_missing_engine


@pytest.fixture
def mca_model(mock_data_array, dim):
    mca = HilbertMCA(n_modes=5)
    mca.fit(mock_data_array, mock_data_array, dim)
    return mca


@pytest.fixture
def mca_model_delayed(mock_dask_data_array, dim):
    mca = HilbertMCA(n_modes=5)
    mca.fit(mock_dask_data_array, mock_dask_data_array, dim)
    return mca


def test_init():
    mca_rotator = HilbertMCARotator(n_modes=2)
    assert mca_rotator._params["n_modes"] == 2
    assert mca_rotator._params["power"] == 1
    assert mca_rotator._params["max_iter"] == 1000
    assert mca_rotator._params["rtol"] == 1e-8


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit(mca_model):
    mca_rotator = HilbertMCARotator(n_modes=2)
    mca_rotator.fit(mca_model)

    assert hasattr(mca_rotator, "model_data")
    assert hasattr(mca_rotator, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_transform(mca_model, mock_data_array):
    mca_rotator = HilbertMCARotator(n_modes=2)
    mca_rotator.fit(mca_model)

    with pytest.raises(NotImplementedError):
        mca_rotator.transform(X=mock_data_array, Y=mock_data_array)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_inverse_transform(mca_model):
    mca_rotator = HilbertMCARotator(n_modes=2)
    mca_rotator.fit(mca_model)

    scores1 = mca_rotator.data["scores1"].sel(mode=slice(1, 3))
    scores2 = mca_rotator.data["scores2"].sel(mode=slice(1, 3))

    reconstructed_data = mca_rotator.inverse_transform(scores1, scores2)

    assert isinstance(reconstructed_data, list)
    assert len(reconstructed_data) == 2


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_squared_covariance_fraction(mca_model, mock_data_array, dim):
    mca_model.fit(mock_data_array, mock_data_array, dim)
    squared_covariance_fraction = mca_model.squared_covariance_fraction()
    assert isinstance(squared_covariance_fraction, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_covariance_fraction(mca_model):
    mca_rotator = HilbertMCARotator(n_modes=4)
    mca_rotator.fit(mca_model)
    cf = mca_rotator.covariance_fraction_CD95()
    assert isinstance(cf, xr.DataArray)
    assert all(cf <= 1), "Squared covariance fraction is greater than 1"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components(mca_model, mock_data_array, dim):
    mca_rotator = HilbertMCARotator(n_modes=2)
    mca_rotator.fit(mca_model)
    comps1, comps2 = mca_rotator.components()
    assert isinstance(comps1, xr.DataArray)
    assert isinstance(comps2, xr.DataArray)
    # assert that the components are Hilbert valued
    assert np.iscomplexobj(comps1)
    assert np.iscomplexobj(comps2)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores(mca_model, mock_data_array, dim):
    mca_rotator = HilbertMCARotator(n_modes=2)
    mca_rotator.fit(mca_model)
    scores1, scores2 = mca_rotator.scores()
    assert isinstance(scores1, xr.DataArray)
    assert isinstance(scores2, xr.DataArray)
    # assert that the scores are Hilbert valued
    assert np.iscomplexobj(scores1)
    assert np.iscomplexobj(scores2)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_homogeneous_patterns(mca_model, mock_data_array, dim):
    mca_rotator = HilbertMCARotator(n_modes=2)
    mca_rotator.fit(mca_model)
    mca_rotator.homogeneous_patterns()


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_heterogeneous_patterns(mca_model, mock_data_array, dim):
    mca_rotator = HilbertMCARotator(n_modes=2)
    mca_rotator.fit(mca_model)
    mca_rotator.heterogeneous_patterns()


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_amplitude(mca_model, mock_data_array, dim):
    mca_rotator = HilbertMCARotator(n_modes=2)
    mca_rotator.fit(mca_model)
    amps1, amps2 = mca_rotator.components_amplitude()


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_phase(mca_model, mock_data_array, dim):
    mca_rotator = HilbertMCARotator(n_modes=2)
    mca_rotator.fit(mca_model)
    amps1, amps2 = mca_rotator.components_phase()


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_amplitude(mca_model, mock_data_array, dim):
    mca_rotator = HilbertMCARotator(n_modes=2)
    mca_rotator.fit(mca_model)
    amps1, amps2 = mca_rotator.scores_amplitude()


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_phase(mca_model, mock_data_array, dim):
    mca_rotator = HilbertMCARotator(n_modes=2)
    mca_rotator.fit(mca_model)
    amps1, amps2 = mca_rotator.scores_phase()


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
# Currently, netCDF4 does not support complex numbers, so skip this test
@pytest.mark.parametrize("engine", ["h5netcdf", "zarr"])
def test_save_load_with_data(tmp_path, engine, mca_model):
    """Test save/load methods in HilbertMCARotator class, ensuring that we can
    roundtrip the model and get the same results."""
    skip_if_missing_engine(engine)

    original = HilbertMCARotator(n_modes=2)
    original.fit(mca_model)

    # Save the HilbertMCARotator model
    original.save(tmp_path / "mca", engine=engine, save_data=True)

    # Check that the HilbertMCARotator model has been saved
    assert (tmp_path / "mca").exists()

    # Recreate the model from saved file
    loaded = HilbertMCARotator.load(tmp_path / "mca", engine=engine)

    # Check that the params and DataContainer objects match
    assert original.get_params() == loaded.get_params()
    assert all([key in loaded.data for key in original.data])
    for key in original.data:
        assert loaded.data[key].equals(original.data[key])

    # Test that the recreated model can compute the SCF
    assert np.allclose(
        original.squared_covariance_fraction(), loaded.squared_covariance_fraction()
    )

    # Test that the recreated model can compute the components amplitude
    A1_original, A2_original = original.components_amplitude()
    A1_loaded, A2_loaded = loaded.components_amplitude()
    assert np.allclose(A1_original, A1_loaded)
    assert np.allclose(A2_original, A2_loaded)

    # Test that the recreated model can compute the components phase
    P1_original, P2_original = original.components_phase()
    P1_loaded, P2_loaded = loaded.components_phase()
    assert np.allclose(P1_original, P1_loaded)
    assert np.allclose(P2_original, P2_loaded)



import numpy as np
import xarray as xr
from scipy.signal import hilbert  # type: ignore
from .data_types import DataArray


def hilbert_transform(
    data: DataArray, dims, padding: str = "exp", decay_factor: float = 0.2
) -> DataArray:
    """Hilbert transform with optional padding to mitigate spectral leakage.

    Parameters:
    ------------
    data: DataArray
        Input data.
    dim: str
        Dimension along which to apply the Hilbert transform.
    padding: str
        Padding type. Can be 'exp' or None.
    decay_factor: float
        Decay factor of the exponential function.

    Returns:
    ---------
    data: DataArray
        Hilbert transform of the input data.

    """
    return xr.apply_ufunc(
        _hilbert_transform_with_padding,
        data,
        input_core_dims=[dims],
        output_core_dims=[dims],
        kwargs={"padding": padding, "decay_factor": decay_factor},
        dask="parallelized",
        dask_gufunc_kwargs={"allow_rechunk": True},
    )


def _hilbert_transform_with_padding(y, padding: str = "exp", decay_factor: float = 0.2):
    """Hilbert transform with optional padding to mitigate spectral leakage.

    Parameters:
    ------------
    y: np.ndarray
        Input array.
    padding: str
        Padding type. Can be 'exp' or None.
    decay_factor: float
        Decay factor of the exponential function.

    Returns:
    ---------
    y: np.ndarray
        Hilbert transform of the input array.

    """
    n_samples = y.shape[0]

    if padding == "exp":
        y = _pad_exp(y, decay_factor=decay_factor)

    y = hilbert(y, axis=0)

    if padding == "exp":
        y = y[n_samples : 2 * n_samples]

    # Padding can introduce a shift in the mean of the imaginary part
    # of the Hilbert transform. Correct for this shift.
    y = y - y.mean(axis=0)  # type: ignore

    return y


def _pad_exp(y, decay_factor: float = 0.2):
    """Pad the input array with an exponential decay function.

    The start and end of the input array are padded with an exponential decay
    function falling to a reference line given by a linear fit of the data array.

    Parameters:
    ------------
    y: np.ndarray
        Input array.
    decay_factor: float
        Decay factor of the exponential function.

    Returns:
    ---------
    y_ext: np.ndarray
        Padded array.

    """
    x = np.arange(y.shape[0])
    x_ext = np.arange(-x.size, 2 * x.size)

    coefs = np.polynomial.polynomial.polyfit(x, y, deg=1)
    yfit = np.polynomial.polynomial.polyval(x, coefs).T
    yfit_ext = np.polynomial.polynomial.polyval(x_ext, coefs).T

    y_ano = y - yfit

    amp_pre = np.take(y_ano, 0, axis=0)[:, None]
    amp_pos = np.take(y_ano, -1, axis=0)[:, None]

    exp_ext = np.exp(-x / x.size / decay_factor)
    exp_ext_reverse = exp_ext[::-1]

    pad_pre = amp_pre * exp_ext_reverse
    pad_pos = amp_pos * exp_ext

    y_ext = np.concatenate([pad_pre.T, y_ano, pad_pos.T], axis=0)
    y_ext += yfit_ext
    return y_ext



from typing_extensions import Self

from ..utils.data_types import Data, DataArray, DataVarBound, Dims
from .transformer import Transformer


class DimensionRenamer(Transformer):
    """Rename dimensions of a DataArray or Dataset.

    Parameters
    ----------
    base: str
        Base string for the new dimension names.
    start: int
        Start index for the new dimension names.

    """

    def __init__(self, base="dim", start=0):
        super().__init__()
        self.base = base
        self.start = start
        self.dim_mapping = {}

    def get_serialization_attrs(self) -> dict:
        return dict(
            dim_mapping=self.dim_mapping,
        )

    def fit(self, X: Data, sample_dims: Dims, feature_dims: Dims, **kwargs) -> Self:
        self.sample_dims_before = sample_dims
        self.feature_dims_before = feature_dims

        self.dim_mapping = {
            dim: f"{self.base}{i}" for i, dim in enumerate(X.dims, start=self.start)
        }

        self.sample_dims_after: Dims = tuple(
            [self.dim_mapping[dim] for dim in self.sample_dims_before]
        )
        self.feature_dims_after: Dims = tuple(
            [self.dim_mapping[dim] for dim in self.feature_dims_before]
        )

        return self

    def transform(self, X: DataVarBound) -> DataVarBound:
        try:
            return X.rename(self.dim_mapping)
        except ValueError:
            raise ValueError("Cannot transform data. Dimensions are different.")

    def _inverse_transform(self, X: DataVarBound) -> DataVarBound:
        given_dims = set(X.dims)
        expected_dims = set(self.dim_mapping.values())
        dims = given_dims.intersection(expected_dims)
        return X.rename({v: k for k, v in self.dim_mapping.items() if v in dims})

    def inverse_transform_data(self, X: DataVarBound) -> DataVarBound:
        return self._inverse_transform(X)

    def inverse_transform_components(self, X: DataVarBound) -> DataVarBound:
        return self._inverse_transform(X)

    def inverse_transform_scores(self, X: DataArray) -> DataArray:
        return self._inverse_transform(X)

    def inverse_transform_scores_unseen(self, X: DataArray) -> DataArray:
        return self._inverse_transform(X)



import numpy as np
import xarray as xr
from typing_extensions import Self

from ..data_container import DataContainer
from ..utils.data_types import DataArray
from .eof import EOF


class ExtendedEOF(EOF):
    """Extended EOF analysis.

    Extended EOF (EEOF) analysis [1]_ [2]_, often referred to as
    Multivariate/Multichannel Singular Spectrum Analysis, enhances
    traditional EOF analysis by identifying propagating signals or
    oscillations in multivariate datasets. This approach integrates the
    spatial correlation of EOFs with the temporal auto- and cross-correlation
    derived from the lagged covariance matrix.

    Parameters
    ----------
    n_modes : int
        Number of modes to be computed.
    tau : int
        Time delay used to construct a time-delayed version of the original time series.
    embedding : int
        Embedding dimension is the number of dimensions in the delay-coordinate space used to represent
        the dynamics of the system. It determines the number of delayed copies
        of the time series that are used to construct the delay-coordinate space.
    n_pca_modes : int, optional
        If provided, the input data is first preprocessed using PCA with the
        specified number of modes. The EEOF analysis is then performed on the
        resulting PCA scores. This approach can lead to important computational
        savings.
    **kwargs :
        Additional keyword arguments passed to the EOF model.

    References
    ----------
    .. [1] Weare, B. C. & Nasstrom, J. S. Examples of Extended Empirical Orthogonal Function Analyses. Monthly Weather Review 110, 481–485 (1982).
    .. [2] Broomhead, D. S. & King, G. P. Extracting qualitative dynamics from experimental data. Physica D: Nonlinear Phenomena 20, 217–236 (1986).


    Examples
    --------
    >>> from xeofs.single import EEOF
    >>> model = EEOF(n_modes=5, tau=1, embedding=20, n_pca_modes=20)
    >>> model.fit(data, dim=("time"))

    Retrieve the extended empirical orthogonal functions (EEOFs) and their explained variance:

    >>> eeofs = model.components()
    >>> exp_var = model.explained_variance()

    Retrieve the time-dependent coefficients corresponding to the EEOF modes:

    >>> scores = model.scores()
    """

    def __init__(
        self,
        n_modes: int,
        tau: int,
        embedding: int,
        n_pca_modes: int | None = None,
        center: bool = True,
        standardize: bool = False,
        use_coslat: bool = False,
        check_nans: bool = True,
        sample_name: str = "sample",
        feature_name: str = "feature",
        compute: bool = True,
        solver: str = "auto",
        random_state: int | None = None,
        solver_kwargs: dict = {},
        **kwargs,
    ):
        super().__init__(
            n_modes=n_modes,
            center=center,
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            sample_name=sample_name,
            feature_name=feature_name,
            compute=compute,
            solver=solver,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
            **kwargs,
        )
        self.attrs.update({"model": "Extended EOF Analysis"})
        self._params.update(
            {"tau": tau, "embedding": embedding, "n_pca_modes": n_pca_modes}
        )

        # Initialize the DataContainer to store the results
        self.data = DataContainer()
        self.pca = (
            EOF(
                n_modes=n_pca_modes,
                center=True,
                standardize=False,
                use_coslat=False,
                compute=self._params["compute"],
                check_nans=False,
                sample_name=self.sample_name,
                feature_name=self.feature_name,
                solver_kwargs=self._params["solver_kwargs"],
            )
            if n_pca_modes
            else None
        )

    def _fit_algorithm(self, X: DataArray) -> Self:
        self.data.add(X.copy(), "input_data", allow_compute=False)

        # Preprocess the data using PCA
        if self.pca:
            self.pca.fit(X, dim=self.sample_name)
            X = self.pca.data["scores"]
            X = X.rename({"mode": self.feature_name})

        # Construct the time-delayed version of the original time series
        tau = self._params["tau"]
        embedding = self._params["embedding"]
        shift = np.arange(embedding) * tau
        X_extended = []
        for i in shift:
            X_extended.append(X.shift(sample=-i))
        X_extended = xr.concat(X_extended, dim="embedding")
        n_samples_cut = (embedding - 1) * tau
        X_extended = X_extended.isel(sample=slice(None, -n_samples_cut))
        X_extended.coords.update({"embedding": shift})

        # Perform standard PCA on extended data
        n_modes = self._params["n_modes"]
        model = EOF(
            n_modes=n_modes,
            center=True,
            standardize=False,
            use_coslat=False,
            compute=self._params["compute"],
            check_nans=False,
            sample_name=self.sample_name,
            feature_name=self.feature_name,
            solver=self._params["solver"],
            solver_kwargs=self._params["solver_kwargs"],
        )
        model.fit(X_extended, dim=self.sample_name)

        self.model = model
        self.data = model.data
        self.data["components"] = model.components()
        self.data["scores"] = model.scores(normalized=False)

        if self.pca:
            self.data["components"] = xr.dot(
                self.pca.data["components"].rename({"mode": "temp"}),
                self.data["components"].rename({"feature": "temp"}),
                dims="temp",
            )

        self.data.set_attrs(self.attrs)

        return self

    def _transform_algorithm(self, X):
        raise NotImplementedError("EEOF does currently not support transform")

    def _inverse_transform_algorithm(self, scores):
        # Reconstruct the data
        comps = self.data["components"].sel(mode=scores.mode, embedding=0, drop=True)

        reconstructed_data = xr.dot(comps.conj(), scores, optimize=True)
        reconstructed_data.name = "reconstructed_data"

        # Enforce real output
        reconstructed_data = reconstructed_data.real

        return reconstructed_data



import dask.array as da
import numpy as np
import pytest
import xarray as xr

from xeofs.cross import HilbertCPCCA, HilbertCPCCARotator


def generate_random_data(shape, lazy=False, seed=142):
    rng = np.random.default_rng(seed)
    if lazy:
        return xr.DataArray(
            da.random.random(shape, chunks=(5, 5)),  # type: ignore
            dims=["sample", "feature"],
            coords={"sample": np.arange(shape[0]), "feature": np.arange(shape[1])},
        )
    else:
        return xr.DataArray(
            rng.random(shape),
            dims=["sample", "feature"],
            coords={"sample": np.arange(shape[0]), "feature": np.arange(shape[1])},
        )


def test_transform_raises():
    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    cpcca = HilbertCPCCA(n_modes=10, alpha=1, use_pca=False)
    cpcca.fit(X, Y, "sample")

    rotator = HilbertCPCCARotator(n_modes=4)
    rotator.fit(cpcca)

    with pytest.raises(NotImplementedError):
        rotator.transform()


@pytest.mark.parametrize(
    "alpha,use_pca",
    [
        (1.0, False),
        (0.5, False),
        (0.0, False),
        (1.0, True),
        (0.5, True),
        (0.0, True),
    ],
)
def test_squared_covariance_fraction_conserved(alpha, use_pca):
    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    cpcca = HilbertCPCCA(n_modes=10, alpha=alpha, use_pca=use_pca, n_pca_modes="all")
    cpcca.fit(X, Y, "sample")

    n_rot_modes = 5
    rotator = HilbertCPCCARotator(n_modes=n_rot_modes, power=1)
    rotator.fit(cpcca)

    scf = rotator.squared_covariance_fraction()
    scf_rot = rotator.squared_covariance_fraction()

    scf_sum = scf.sel(mode=slice(1, n_rot_modes)).sum()
    scf_rot_sum = scf_rot.sel(mode=slice(1, n_rot_modes)).sum()

    xr.testing.assert_allclose(scf_sum, scf_rot_sum)



import numpy as np
import numba
from numba import prange
from scipy.spatial.distance import cdist

from ..constants import AVG_EARTH_RADIUS


def distance_matrix_bc(A, B, metric="haversine"):
    """Compute a distance matrix between two arrays using broadcasting.

    Parameters
    ----------
    A: 2D darray
        Array of longitudes and latitudes with shape (N, 2)
    B: 2D darray
        Array of longitudes and latitudes with shape (M, 2)
    metric: str
        Distance metric to use. Great circle distance (`haversine`) is always expressed in kilometers.
        All other distance metrics are reported in the unit of the input data.
        See scipy.spatial.distance.cdist for a list of available metrics.

    Returns
    -------
    distance: 2D darray
        Distance matrix with shape (N, M)


    """
    if metric == "haversine":
        return _haversine_distance_bc(A, B)
    else:
        return cdist(XA=A, XB=B, metric=metric)


def _haversine_distance_bc(lonlats1, lonlats2):
    """Compute the great circle distance matrix between two arrays

    This implementation uses numpy broadcasting.

    Parameters
    ----------
    lonlats1: 2D darray
        Array of longitudes and latitudes with shape (N, 2)
    lonlats2: 2D darray
        Array of longitudes and latitudes with shape (M, 2)

    Returns
    -------
    distance: 2D darray
        Great circle distance matrix with shape (N, M) in kilometers

    """
    # Convert to radians
    lonlats1 = np.radians(lonlats1)
    lonlats2 = np.radians(lonlats2)

    # Extract longitudes and latitudes
    lon1, lat1 = lonlats1[:, 0], lonlats1[:, 1]
    lon2, lat2 = lonlats2[:, 0], lonlats2[:, 1]

    # Compute differences in longitudes and latitudes
    dlon = lon2 - lon1[:, np.newaxis]
    dlat = lat2 - lat1[:, np.newaxis]

    # Haversine formula
    a = (
        np.sin(dlat / 2) ** 2
        + np.cos(lat1)[..., None] * np.cos(lat2) * np.sin(dlon / 2) ** 2
    )
    c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))
    distance = AVG_EARTH_RADIUS * c

    return distance


@numba.njit(fastmath=True)
def distance_nb(A, b, metric="euclidean"):
    if metric == "euclidean":
        return _euclidian_distance_nb(A, b)
    elif metric == "haversine":
        return _haversine_distance_nb(A, b)
    else:
        raise ValueError(
            f"Invalid metric: {metric}. Must be one of ['euclidean', 'haversine']."
        )


@numba.njit(fastmath=True)
def _euclidian_distance_nb(A, b):
    """Compute the Euclidian distance between two arrays.

    This implementation uses numba.

    Parameters
    ----------
    A: 2D array
        Array of shape (N, P)
    b: 1D array
        Array of shape (P,)

    Returns
    -------
    distance: 1D array
        Distance matrix with shape (N,)

    """
    dist = np.zeros(A.shape[0])
    for r in prange(A.shape[0]):
        d = 0
        for c in range(A.shape[1]):
            d += (b[c] - A[r, c]) ** 2
        dist[r] = d
    return np.sqrt(dist)


@numba.njit(fastmath=True)
def _haversine_distance_nb(A, b):
    # Convert to radians
    A = np.radians(A)
    b = np.radians(b)

    # Compute differences in longitudes and latitudes
    dlon = b[0] - A[:, 0]
    dlat = b[1] - A[:, 1]

    # Haversine formula
    a = np.sin(dlat / 2) ** 2 + np.cos(A[:, 1]) * np.cos(b[1]) * np.sin(dlon / 2) ** 2
    c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))
    distance = AVG_EARTH_RADIUS * c

    return distance



from ast import literal_eval
from typing import Any

import numpy as np
import xarray as xr


def write_model_tree(
    dt: xr.DataTree, path: str, overwrite: bool = False, engine: str = "zarr", **kwargs
):
    """Write a DataTree to a file."""
    write_mode = "w" if overwrite else "w-"
    if engine in ["netcdf4", "h5netcdf"]:
        dt = _sanitize_attrs_nc(dt)
        dt.to_netcdf(path, engine=engine, **kwargs)
    elif engine == "zarr":
        dt.to_zarr(path, mode=write_mode, **kwargs)
    else:
        raise ValueError(f"Unknown engine {engine}")


def open_model_tree(path: str, engine: str = "zarr", **kwargs) -> xr.DataTree:
    """Open a DataTree from a file."""
    if engine == "zarr" and "chunks" not in kwargs:
        kwargs["chunks"] = {}
    dt = xr.open_datatree(path, engine=engine, **kwargs)
    if engine in ["netcdf4", "h5netcdf"]:
        dt = _desanitize_attrs_nc(dt)
    return dt


def insert_placeholders(dt: xr.DataTree) -> xr.DataTree:
    """Insert placeholders for data that we don't want to compute."""
    for node in dt.subtree:
        if not node.attrs.get("allow_compute", True):
            dt[node.path] = xr.DataTree(
                xr.Dataset(
                    data_vars={
                        node.name: xr.DataArray(np.nan, attrs={"placeholder": True})
                    },
                    attrs={"allow_compute": False, "placeholder": True},
                )
            )
    return dt


def _sanitize_attrs_nc(dt: xr.DataTree) -> xr.DataTree:
    """Sanitize both node-level and variable-level attrs to strings for netcdf."""
    sanitized_types = (dict, list, bool, type(None))
    for node in dt.subtree:
        for key, attr in node.attrs.items():
            if isinstance(attr, sanitized_types):
                node.attrs[key] = str(attr)
        for v in node.variables:
            for key, attr in node[v].attrs.items():
                if isinstance(attr, sanitized_types):
                    node[v].attrs[key] = str(attr)
    return dt


def _should_desanitize(attr: Any) -> bool:
    if isinstance(attr, str):
        if (
            (attr[0] == "{" and attr[-1] == "}")
            or (attr[0] == "[" and attr[-1] == "]")
            or (attr in ["True", "False"])
            or (attr == "None")
        ):
            return True
    return False


def _desanitize_attrs_nc(dt: xr.DataTree) -> xr.DataTree:
    """Desanitize both node-level and variable-level attrs from strings for netcdf."""
    for node in dt.subtree:
        for key, attr in node.attrs.items():
            if _should_desanitize(attr):
                node.attrs[key] = literal_eval(attr)
        for v in node.variables:
            for key, attr in node[v].attrs.items():
                if _should_desanitize(attr):
                    node[v].attrs[key] = literal_eval(attr)
    return dt



import numpy as np

import numba
from numba import prange
from ..optional.distance_metrics import distance_nb
from ..optional.kernels import kernel_weights_nb


# Additional utility functions for local PCA
# =============================================================================


@numba.njit(fastmath=True, parallel=True)
def _local_pcas(X, xy, n_modes, metric, kernel, bandwidth):
    """Perform local PCA on each sample.

    Parameters
    ----------
    X: ndarray
        Input data with shape (n_samples, n_features)
    xy: ndarray
        Sample coordinates with shape (n_samples, 2)
    n_modes: int
        Number of modes to calculate.
    metric: str
        Distance metric to use. Great circle distance (`haversine`) is always expressed in kilometers.
        All other distance metrics are reported in the unit of the input data.
        See scipy.spatial.distance.cdist for a list of available metrics.
    kernel: str
        Kernel function to use. Must be one of ['bisquare', 'gaussian', 'exponential'].
    bandwidth: float
        Bandwidth of the kernel function.

    Returns
    -------
    ndarray
        Array of local components with shape (n_samples, n_features, n_modes)
    ndarray
        Array of local explained variance with shape (n_samples, n_modes)
    ndarray
        Array of total variance with shape (n_samples,)

    """
    n_samples = X.shape[0]
    n_features = X.shape[1]
    Vs = np.empty((n_samples, n_features, n_modes))
    exp_var = np.empty((n_samples, n_modes))
    tot_var = np.empty(n_samples)
    for i in prange(n_samples):
        dist = distance_nb(xy, xy[i], metric=metric)
        weights = kernel_weights_nb(dist, bandwidth, kernel)
        valid_data = weights > 0

        weights = weights[valid_data]
        x = X[valid_data]

        wmean = _wmean_axis0(x, weights)
        x -= wmean

        sqrt_weights = np.sqrt(weights)
        x = _weigh_columns(x, sqrt_weights)

        Ui, si, ViT = np.linalg.svd(x, full_matrices=False)
        # Renormalize singular values
        si = si**2 / weights.sum()
        ti = si.sum()

        si = si[:n_modes]
        ViT = ViT[:n_modes]
        Vi = ViT.T

        Vs[i] = Vi
        exp_var[i, : len(si)] = si
        tot_var[i] = ti

    return Vs, exp_var, tot_var


@numba.njit(fastmath=True)
def _wmean_axis0(X, weights):
    """Compute weighted mean along axis 0.

    Numba version of np.average. Note that np.average is supported by Numba,
    but is restricted to `X` and `weights` having the same shape.
    """
    wmean = np.empty(X.shape[1])
    for i in prange(X.shape[1]):
        wmean[i] = np.average(X[:, i], weights=weights)
    return wmean


@numba.njit(fastmath=True)
def _weigh_columns(x, weights):
    """Weigh columns of x by weights.

    Numba version of broadcasting.

    Parameters
    ----------
    x: ndarray
        Input data with shape (n_samples, n_features)
    weights: ndarray
        Weights with shape (n_samples,)

    Returns
    -------
    x_weighted: ndarray
        Weighted data with shape (n_samples, n_features)
    """
    x_weighted = np.zeros_like(x)
    for i in range(x.shape[1]):
        x_weighted[:, i] = x[:, i] * weights
    return x_weighted


@numba.guvectorize(
    [
        (
            numba.float32[:, :],
            numba.float32[:, :],
            numba.float32[:],
            numba.int32[:],
            numba.int32,
            numba.float32,
            numba.int32,
            numba.float32[:, :],
            numba.float32[:],
            numba.float32,
        )
    ],
    # In order to specify the output dimension which has not been defined in the input dimensions
    # one has to use a dummy variable (see Numba #2797 https://github.com/numba/numba/issues/2797)
    "(n,m),(n,o),(o),(n_out),(),(),()->(m,n_out),(n_out),()",
)
def local_pca_vectorized(
    data, XY, xy, n_out, metric, bandwidth, kernel, comps, expvar, totvar
):
    """Perform local PCA

    Numba vectorized version of local_pca.

    Parameters
    ----------
    data: ndarray
        Input data with shape (n_samples, n_features)
    XY: ndarray
        Sample coordinates with shape (n_samples, 2)
    xy: ndarray
        Coordinates of the sample to perform PCA on with shape (2,)
    n_out: ndarray
        Number of modes to calculate. (see comment above; workaround for Numba #2797)
    metric: int
        Numba only accepts int/floats; so metric str has to be converted first e.g. by a simple dictionary (not implemented yet)
        see Numba #4404 (https://github.com/numba/numba/issues/4404)
    bandwidth: float
        Bandwidth of the kernel function.
    kernel: int
        Numba only accepts int/floats; so kernel str has to be converted first e.g. by a simple dictionary (not implemented yet)
        see Numba #4404 (https://github.com/numba/numba/issues/4404)
    comps: ndarray
        Array of local components with shape (n_features, n_modes)
    expvar: ndarray
        Array of local explained variance with shape (n_modes)
    totvar: ndarray
        Array of total variance with shape (1)


    """
    distance = distance_nb(XY, xy, metric=metric)
    weights = kernel_weights_nb(distance, bandwidth, kernel)
    is_positive_weight = weights > 0
    X = data[is_positive_weight]
    weights = weights[is_positive_weight]

    wmean = _wmean_axis0(X, weights)
    X -= wmean

    sqrt_weights = np.sqrt(weights)
    X = _weigh_columns(X, sqrt_weights)

    U, s, Vt = np.linalg.svd(X, full_matrices=False)
    Vt = Vt[: n_out.shape[0], :]
    lbda = s**2 / weights.sum()
    for i in range(n_out.shape[0]):
        expvar[i] = lbda[i]
        comps[:, i] = Vt[i, :]
    lbda.sum()



"""
Extented EOF analysis
=====================

This example demonstrates Extended EOF (EEOF) analysis on ``xarray`` tutorial
data. EEOF analysis, also termed as Multivariate/Multichannel Singular
Spectrum Analysis, advances traditional EOF analysis to capture propagating
signals or oscillations in multivariate datasets. At its core, this
involves the formulation of a lagged covariance matrix that encapsulates
both spatial and temporal correlations. Subsequently, this matrix is
decomposed to yield its eigenvectors (components) and eigenvalues (explained variance).

Let's begin by setting up the required packages and fetching the data:
"""

import matplotlib.pyplot as plt
import xarray as xr

import xeofs as xe

xr.set_options(display_expand_data=False)

# %%
# Load the tutorial data.
t2m = xr.tutorial.load_dataset("air_temperature").air


# %%
# Prior to conducting the EEOF analysis, it's essential to determine the
# structure of the lagged covariance matrix. This entails defining the time
# delay ``tau`` and the ``embedding`` dimension. The former signifies the
# interval between the original and lagged time series, while the latter
# dictates the number of time-lagged copies in the delay-coordinate space,
# representing the system's dynamics.
# For illustration, using ``tau=4`` and ``embedding=40``, we generate 40
# delayed versions of the time series, each offset by 4 time steps, resulting
# in a maximum shift of ``tau x embedding = 160``. Given our dataset's
# 6-hour intervals, tau = 4 translates to a 24-hour shift.
# It's obvious that this way of constructing the lagged covariance matrix
# and subsequently decomposing it can be computationally expensive. For example,
# given our dataset's dimensions,

t2m.shape

# %%
# the extended dataset would have 40 x 25 x 53 = 53000 features
# which is much larger than the original dataset's 1325 features.
# To mitigate this, we can first preprocess the data using PCA / EOF analysis
# and then perform EEOF analysis on the resulting PCA / EOF scores. Here,
# we'll use ``n_pca_modes=50`` to retain the first 50 PCA modes, so we end
# up with 40 x 50 = 200 (latent) features.
# With these parameters set, we proceed to instantiate the ``ExtendedEOF``
# model and fit our data.

model = xe.single.ExtendedEOF(
    n_modes=10, tau=4, embedding=40, n_pca_modes=50, use_coslat=True
)
model.fit(t2m, dim="time")
scores = model.scores()
components = model.components()
components

# %%
# A notable distinction from standard EOF analysis is the incorporation of an
# extra ``embedding`` dimension in the components. Nonetheless, the
# overarching methodology mirrors traditional EOF practices. The results,
# for instance, can be assessed by examining the explained variance ratio.

model.explained_variance_ratio().plot()
plt.show()

# %%
# Additionally, we can look into the scores; let's spotlight mode 4.

scores.sel(mode=4).plot()
plt.show()

# %%
# In wrapping up, we visualize the corresponding EEOF component of mode 4.
# For visualization purposes, we'll focus on the component at a specific
# latitude, in this instance, 60 degrees north.

components.sel(mode=4, lat=60).plot()
plt.show()



"""
Canonical Correlation Analysis
==============================

In this example, we're going to perform a Canonical Correlation Analysis (CCA)
on three datasets using the ERSSTv5 monthly sea surface temperature (SST) data
from 1970 to 2022. We divide this data into three areas: the Indian Ocean,
the Pacific Ocean, and the Atlantic Ocean. Our goal is to perform CCA on these
regions.

First, we'll import the necessary modules.
"""

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import xarray as xr
from matplotlib.gridspec import GridSpec

import xeofs as xe

# %%
# Next, we load the data and compute the SST anomalies. This removes the
# monthly climatologies, so the seasonal cycle doesn't impact our CCA.

sst = xr.tutorial.load_dataset("ersstv5").sst
sst = sst.groupby("time.month") - sst.groupby("time.month").mean("time")


# %%
# Now, we define the three regions of interest and store them in a list.

indian = sst.sel(lon=slice(35, 115), lat=slice(30, -30))
pacific = sst.sel(lon=slice(130, 290), lat=slice(30, -30))
atlantic = sst.sel(lon=slice(320, 360), lat=slice(70, 10))

data_list = [indian, pacific, atlantic]

# %%
# We now perform CCA. Since we are dealing with a high-dimensional feature space, we first
# perform PCA to reduce the dimensionality (this is kind of a regularized CCA) by setting
# ``pca=True``. By setting the ``variance_fraction`` keyword argument, we specify that we
# want to keep the number of PCA modes that explain 90% of the variance in each of the
# three data sets.
#
# An important parameter is ``init_pca_modes``. It specifies the number
# of PCA modes that are initially compute before truncating them to account for 90 %. If this
# number is small enough, randomized PCAs will be performed instead of the full SVD decomposition
# which is much faster. We can also specify ``init_pca_modes`` as a float (0 < x <= 1),
# in which case the number of PCA modes is given by the fraction of the data matrix's rank
# The default is set to 0.75 which will ensure that randomized PCAs are performed.
#
# Given the nature of SST data, we might lower it to something like 0.3, since
# we expect that most of the variance in the data will be explained by a small
# number of PC modes.
#
# Note that if our initial PCA modes don't hit the 90% variance target, ``xeofs``
# will give a warning.

model = xe.multi.CCA(
    n_modes=2,
    use_coslat=True,
    pca=True,
    variance_fraction=0.9,
    init_pca_modes=0.30,
)
model.fit(data_list, dim="time")
components = model.components()
scores = model.scores()

# %%
# Let's look at the canonical loadings (components) of the first mode.

mode = 1

central_longitudes = [
    indian.lon.median().item(),
    pacific.lon.median().item(),
    pacific.lon.median().item(),
]
projections = [ccrs.PlateCarree(central_longitude=lon) for lon in central_longitudes]

fig = plt.figure(figsize=(12, 2.5))
gs = GridSpec(1, 4, figure=fig, width_ratios=[2, 4, 1, 0.2])
axes = [fig.add_subplot(gs[0, i], projection=projections[i]) for i in range(3)]
cax = fig.add_subplot(1, 4, 4)
kwargs = dict(transform=ccrs.PlateCarree(), vmin=-1, vmax=1, cmap="RdBu_r", cbar_ax=cax)
components[0].sel(mode=mode).plot(ax=axes[0], **kwargs)
components[1].sel(mode=mode).plot(ax=axes[1], **kwargs)
im = components[2].sel(mode=mode).plot(ax=axes[2], **kwargs)
fig.colorbar(im, cax=cax, orientation="vertical")
for ax in axes:
    ax.coastlines()
    ax.set_title("")

# %%
# And lastly, we'll check out the canonical variates (scores) of the first mode.

fig, ax = plt.subplots(figsize=(12, 4))
scores[0].sel(mode=mode).plot(ax=ax, label="Indian Ocean")
scores[1].sel(mode=mode).plot(ax=ax, label="Central Pacific")
scores[2].sel(mode=mode).plot(ax=ax, label="North Atlantic")
ax.legend()



import numpy as np
import pytest
import xarray as xr

# Import the classes from your modules
from xeofs.cross import MCA, MCARotator

from ...utilities import data_is_dask, skip_if_missing_engine


@pytest.fixture
def mca_model(mock_data_array, dim):
    mca = MCA(n_modes=5, use_pca=False)
    mca.fit(mock_data_array, mock_data_array, dim)
    return mca


@pytest.fixture
def mca_model_delayed(mock_dask_data_array, dim):
    mca = MCA(n_modes=5, compute=False, check_nans=False, use_pca=False)
    mca.fit(mock_dask_data_array, mock_dask_data_array, dim)
    return mca


def test_init():
    mca_rotator = MCARotator(n_modes=4)
    assert mca_rotator._params["n_modes"] == 4
    assert mca_rotator._params["power"] == 1
    assert mca_rotator._params["max_iter"] == 1000
    assert mca_rotator._params["rtol"] == 1e-8
    # assert mca_rotator._params["squared_loadings"] is False


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit(mca_model):
    mca_rotator = MCARotator(n_modes=4)
    mca_rotator.fit(mca_model)

    assert hasattr(mca_rotator, "model_data")
    assert hasattr(mca_rotator, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_transform(mca_model, mock_data_array):
    mca_rotator = MCARotator(n_modes=4)
    mca_rotator.fit(mca_model)

    projections = mca_rotator.transform(X=mock_data_array, Y=mock_data_array)

    assert len(projections) == 2


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_inverse_transform(mca_model):
    mca_rotator = MCARotator(n_modes=4)
    mca_rotator.fit(mca_model)

    scores1 = mca_rotator.data["scores1"].sel(mode=slice(1, 3))
    scores2 = mca_rotator.data["scores2"].sel(mode=slice(1, 3))

    reconstructed_data = mca_rotator.inverse_transform(scores1, scores2)

    assert isinstance(reconstructed_data, list)
    assert len(reconstructed_data) == 2


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_squared_covariance_fraction(mca_model, mock_data_array, dim):
    mca_rotator = MCARotator(n_modes=4)
    mca_rotator.fit(mca_model)
    squared_covariance_fraction = mca_rotator.squared_covariance_fraction()
    assert isinstance(squared_covariance_fraction, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_covariance_fraction(mca_model):
    mca_rotator = MCARotator(n_modes=4)
    mca_rotator.fit(mca_model)
    cf = mca_rotator.covariance_fraction_CD95()
    assert isinstance(cf, xr.DataArray)
    assert all(cf <= 1), "Squared covariance fraction is greater than 1"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components(mca_model, mock_data_array, dim):
    mca_rotator = MCARotator(n_modes=4)
    mca_rotator.fit(mca_model)
    comps1, comps2 = mca_rotator.components()
    assert isinstance(comps1, xr.DataArray)
    assert isinstance(comps2, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores(mca_model, mock_data_array, dim):
    mca_rotator = MCARotator(n_modes=4)
    mca_rotator.fit(mca_model)
    scores1, scores2 = mca_rotator.scores()
    assert isinstance(scores1, xr.DataArray)
    assert isinstance(scores2, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_homogeneous_patterns(mca_model, mock_data_array, dim):
    mca_rotator = MCARotator(n_modes=4)
    mca_rotator.fit(mca_model)
    patterns, pvalues = mca_rotator.homogeneous_patterns()
    assert isinstance(patterns[0], xr.DataArray)
    assert isinstance(patterns[1], xr.DataArray)
    assert isinstance(pvalues[0], xr.DataArray)
    assert isinstance(pvalues[1], xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_heterogeneous_patterns(mca_model, mock_data_array, dim):
    mca_rotator = MCARotator(n_modes=4)
    mca_rotator.fit(mca_model)
    patterns, pvalues = mca_rotator.heterogeneous_patterns()
    assert isinstance(patterns[0], xr.DataArray)
    assert isinstance(patterns[1], xr.DataArray)
    assert isinstance(pvalues[0], xr.DataArray)
    assert isinstance(pvalues[1], xr.DataArray)


@pytest.mark.parametrize(
    "dim, compute",
    [
        (("time",), True),
        (("lat", "lon"), True),
        (("lon", "lat"), True),
        (("time",), False),
        (("lat", "lon"), False),
        (("lon", "lat"), False),
    ],
)
def test_compute(mca_model_delayed, compute):
    """Test the compute method of the MCARotator class."""

    mca_rotator = MCARotator(n_modes=4, compute=compute, max_iter=20, rtol=1e-4)
    mca_rotator.fit(mca_model_delayed)

    if compute:
        assert not data_is_dask(mca_rotator.data["squared_covariance"])
        assert not data_is_dask(mca_rotator.data["components1"])
        assert not data_is_dask(mca_rotator.data["components2"])
        assert not data_is_dask(mca_rotator.data["rotation_matrix"])
        assert not data_is_dask(mca_rotator.data["phi_matrix"])
        assert not data_is_dask(mca_rotator.data["norm1"])
        assert not data_is_dask(mca_rotator.data["norm2"])
        assert not data_is_dask(mca_rotator.data["modes_sign"])

    else:
        assert data_is_dask(mca_rotator.data["squared_covariance"])
        assert data_is_dask(mca_rotator.data["components1"])
        assert data_is_dask(mca_rotator.data["components2"])
        assert data_is_dask(mca_rotator.data["rotation_matrix"])
        assert data_is_dask(mca_rotator.data["phi_matrix"])
        assert data_is_dask(mca_rotator.data["norm1"])
        assert data_is_dask(mca_rotator.data["norm2"])
        assert data_is_dask(mca_rotator.data["modes_sign"])


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
@pytest.mark.parametrize("engine", ["h5netcdf", "netcdf4", "zarr"])
def test_save_load(dim, mock_data_array, tmp_path, engine):
    """Test save/load methods in MCA class, ensuring that we can
    roundtrip the model and get the same results when transforming
    data."""
    skip_if_missing_engine(engine)

    original_unrotated = MCA()
    original_unrotated.fit(mock_data_array, mock_data_array, dim)

    original = MCARotator()
    original.fit(original_unrotated)

    # Save the EOF model
    original.save(tmp_path / "mca", engine=engine)

    # Check that the EOF model has been saved
    assert (tmp_path / "mca").exists()

    # Recreate the model from saved file
    loaded = MCARotator.load(tmp_path / "mca", engine=engine)

    # Check that the params and DataContainer objects match
    assert original.get_params() == loaded.get_params()
    assert all([key in loaded.data for key in original.data])
    for key in original.data:
        if original.data._allow_compute[key]:
            assert loaded.data[key].equals(original.data[key])
        else:
            # but ensure that input data is not saved by default
            assert loaded.data[key].size <= 1
            assert loaded.data[key].attrs["placeholder"] is True

    # Test that the recreated model can be used to transform new data
    assert np.allclose(
        original.transform(mock_data_array, mock_data_array),
        loaded.transform(mock_data_array, mock_data_array),
    )

    # The loaded model should also be able to inverse_transform new data
    assert np.allclose(
        original.inverse_transform(*original.scores()),
        loaded.inverse_transform(*loaded.scores()),
    )


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_serialize_deserialize_dataarray(dim, mock_data_array):
    """Test roundtrip serialization when the model is fit on a DataArray."""
    model = MCA()
    model.fit(mock_data_array, mock_data_array, dim)
    rotator = MCARotator()
    rotator.fit(model)
    dt = rotator.serialize()
    rebuilt_rotator = MCARotator.deserialize(dt)
    assert np.allclose(
        rotator.transform(mock_data_array), rebuilt_rotator.transform(mock_data_array)
    )


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_serialize_deserialize_dataset(dim, mock_dataset):
    """Test roundtrip serialization when the model is fit on a Dataset."""
    model = MCA()
    model.fit(mock_dataset, mock_dataset, dim)
    rotator = MCARotator()
    rotator.fit(model)
    dt = rotator.serialize()
    rebuilt_rotator = MCARotator.deserialize(dt)
    assert np.allclose(
        rotator.transform(mock_dataset), rebuilt_rotator.transform(mock_dataset)
    )



import pandas as pd
import xarray as xr
from typing_extensions import Self

from ..utils.data_types import Data, DataArray, DataSet, DataVar, DataVarBound, Dims
from .transformer import Transformer


class Stacker(Transformer):
    """Converts an xarray DataArray or Dataset of any dimensionality into a 2D DataArray.

    The new DataArray will have two dimensions: `sample` and `feature`.
    The dimensions of the original data will be stacked along these two dimensions.

    Attributes
    ----------
    sample_dims : Sequence[Hashable]
        The dimensions of the data that will be stacked along the `sample` dimension.
    feature_dims : Sequence[Hashable]
        The dimensions of the data that will be stacked along the `feature` dimension.
    sample_name : str
        The name of the sample dimension (dim=0).
    feature_name : str
        The name of the feature dimension (dim=1).
    dims_in : tuple[str]
        The dimensions of the input data.
    dims_out : tuple[str]
        The dimensions of the output data.
    dims_mapping : dict[str, tuple[str]]
        The mapping between the input and output dimensions.
    coords_in : dict[str, xr.Coordinates]
        The coordinates of the input data.
    coords_out : dict[str, xr.Coordinates]
        The coordinates of the output data.
    """

    def __init__(
        self,
        sample_name: str = "sample",
        feature_name: str = "feature",
    ):
        super().__init__(sample_name, feature_name)

        self.dims_in = tuple()
        self.dims_out = tuple((sample_name, feature_name))
        self.dims_mapping = {}
        self.dims_mapping.update({d: tuple() for d in self.dims_out})
        self.coords_in = {}
        self.coords_out = {}
        self.data_type = None

    def get_serialization_attrs(self) -> dict:
        return dict(
            dims_in=self.dims_in,
            dims_out=self.dims_out,
            dims_mapping=self.dims_mapping,
            coords_in=self.coords_in,
            coords_out=self.coords_out,
            data_type=self.data_type,
        )

    def _validate_data_type(self, X: Data):
        """Check that the data type is either DataArray or Dataset."""
        if not isinstance(X, (xr.DataArray, xr.Dataset)):
            raise TypeError(f"Invalid data type {type(X)}.")

    def _validate_dimension_names(self, X, sample_dims, feature_dims):
        """Check that the names of the sample and feature dimensions are not already present in the data."""
        sample_name_in_data = self.sample_name in X.dims
        feature_name_in_data = self.feature_name in X.dims

        has_invalid_sample_name = (
            True if (len(sample_dims) > 1) and sample_name_in_data else False
        )

        match X:
            case xr.DataArray():
                has_invalid_feature_name = (
                    True if (len(feature_dims) > 1) and feature_name_in_data else False
                )
            case xr.Dataset():
                has_invalid_feature_name = True if feature_name_in_data else False
            case _:
                raise TypeError(f"Invalid data type {type(X)}.")

        if has_invalid_sample_name:
            err_msg = f"Name of sample dimension ({self.sample_name}) is already present in data. Please use another name."
            raise ValueError(err_msg)

        if has_invalid_feature_name:
            err_msg = f"Name of feature dimension ({self.feature_name}) is already present in data. Please use another name."
            raise ValueError(err_msg)

    def _validate_dims(self, X: Data, sample_dims, feature_dims):
        invalid_sample_dims = True if len(sample_dims) < 1 else False
        invalid_feature_dims = True if len(feature_dims) < 1 else False

        if invalid_sample_dims:
            raise ValueError("Sample dimension must not be empty.")
        if invalid_feature_dims:
            match X:
                case xr.DataArray():
                    raise ValueError("Feature dimension must not be empty.")
                case xr.Dataset():
                    err_msg = "Dataset without feature dimension is currently not supported. Please convert your Dataset to a DataArray first, e.g. by using `to_array()`."
                    raise ValueError(err_msg)
                case _:
                    raise TypeError(f"Invalid data type {type(X)}.")

    def _validate_indices(self, X: Data):
        """Check that the indices of the data are no MultiIndex"""
        if any([isinstance(index, pd.MultiIndex) for index in X.indexes.values()]):
            raise ValueError("Cannot stack data containing a MultiIndex.")

    def _sanity_check(self, X: Data, sample_dims, feature_dims):
        self._validate_dims(X, sample_dims, feature_dims)
        self._validate_dimension_names(X, sample_dims, feature_dims)
        self._validate_indices(X)

    def _validate_transform_data_type(self, X: Data):
        """Check that the data type is either DataArray or Dataset."""
        if self._type_name(X) != self.data_type:
            raise TypeError(f"Expected data type {self.data_type}, got {type(X)}.")

    def _validate_transform_dimensions(self, X: Data):
        """Verify that the dimensions of the data are consistent with the dimensions used to fit the stacker."""
        # Test whether sample and feature dimensions are present in data array
        expected_sample_dims = set(self.dims_mapping[self.sample_name])
        expected_feature_dims = set(self.dims_mapping[self.feature_name])
        expected_dims = expected_sample_dims | expected_feature_dims
        given_dims = set(X.dims)
        if not (expected_dims == given_dims):
            raise ValueError(
                f"One or more dimensions in {expected_dims} are not present in data."
            )

    def _validate_transform_feature_coords(self, X: Data):
        """Verify that the feature coordinates of the data are consistent with the feature coordinates used to fit the stacker."""
        feature_dims = self.dims_mapping[self.feature_name]
        coords_are_equal = [
            X.coords[dim].equals(self.coords_in[dim]) for dim in feature_dims
        ]
        if not all(coords_are_equal):
            raise ValueError(
                "Data to be transformed has different coordinates than the data used to fit."
            )

    def _reorder_dims(self, X: DataVarBound) -> DataVarBound:
        """Reorder dimensions to original order; catch ('mode') dimensions via ellipsis"""
        order_input_dims = [
            valid_dim for valid_dim in self.dims_in if valid_dim in X.dims
        ]
        if order_input_dims != X.dims:
            X = X.transpose(..., *order_input_dims)
        return X

    def _stack(self, X: Data, sample_dims: Dims, feature_dims: Dims) -> DataArray:
        """Stack data to 2D.

        Parameters
        ----------
        X : xr.DataArray | xr.Dataset
            The data to be reshaped.
        sample_dims : Hashable or Sequence[Hashable]
            The dimensions of the data that will be stacked along the `sample` dimension.
        feature_dims : Hashable or Sequence[Hashable]
            The dimensions of the data that will be stacked along the `feature` dimension.

        Returns
        -------
        DataArray
            The reshaped 2d-data.
        """
        sample_name = self.sample_name
        feature_name = self.feature_name

        # Stack SAMPLE dimension
        if len(sample_dims) > 1:
            X = X.stack({sample_name: sample_dims})
        elif len(sample_dims) == 1:
            if sample_dims[0] != sample_name:
                X = X.rename({sample_dims[0]: sample_name})
            else:
                # There's only one sample dimension and it's already named correctly
                pass
        else:
            raise ValueError("Sample dimension must not be empty.")

        # Stack FEATURE dimension
        match X:
            case xr.DataArray():
                if len(feature_dims) > 1:
                    X = X.stack({feature_name: feature_dims})
                elif len(feature_dims) == 1:
                    if feature_dims[0] != feature_name:
                        X = X.rename({feature_dims[0]: feature_name})
                    else:
                        # There's only one feature dimension and it's already named correctly
                        pass
                else:
                    raise ValueError("Feature dimension must not be empty.")

            case xr.Dataset():
                X = X.to_stacked_array(
                    new_dim=feature_name, sample_dims=(self.sample_name,)
                )
            case _:
                raise TypeError(f"Invalid data type {type(X)}.")

        # Reorder dimensions to be always (sample, feature)
        if X.dims == (feature_name, sample_name):
            X = X.transpose(sample_name, feature_name)

        return X

    def _unstack_to_dataarray(self, X: DataArray) -> DataArray:
        """Unstack 2D DataArray to its original dimensions.

        Parameters
        ----------
        X : DataArray
            The data to be unstacked.

        Returns
        -------
        DataArray
            The unstacked data.
        """
        sample_name = self.sample_name
        feature_name = self.feature_name

        has_only_one_sample_dim = len(self.dims_mapping[sample_name]) == 1
        has_only_one_feature_dim = len(self.dims_mapping[feature_name]) == 1

        if sample_name in X.dims:
            # If sample dimensions is one dimensional, rename is sufficient, otherwise unstack
            if has_only_one_sample_dim:
                if self.dims_mapping[sample_name][0] != sample_name:
                    X = X.rename({sample_name: self.dims_mapping[sample_name][0]})
            else:
                X = X.unstack(sample_name)

        # pass if feature/sample dimensions do not exist in data
        if feature_name in X.dims:
            # If sample dimensions is one dimensional, rename is sufficient, otherwise unstack
            if has_only_one_feature_dim:
                if self.dims_mapping[feature_name][0] != feature_name:
                    X = X.rename({feature_name: self.dims_mapping[feature_name][0]})
            else:
                X = X.unstack(feature_name)

        else:
            pass

        X = self._reorder_dims(X)
        return X

    def _unstack_to_dataset_data(self, X: DataArray) -> DataSet:
        """Unstack `sample` and `feature` dimension of an DataArray to its original dimensions."""
        sample_name = self.sample_name
        feature_name = self.feature_name
        has_only_one_sample_dim = len(self.dims_mapping[sample_name]) == 1

        if has_only_one_sample_dim:
            X = X.rename({sample_name: self.dims_mapping[sample_name][0]})

        ds: DataSet = X.to_unstacked_dataset(feature_name, "variable").unstack()
        ds = self._reorder_dims(ds)
        return ds

    def _unstack_to_dataset_components(self, data: DataArray) -> DataSet:
        feature_name = self.feature_name
        ds: DataSet = data.to_unstacked_dataset(feature_name, "variable").unstack()
        ds = self._reorder_dims(ds)
        return ds

    def _type_name(self, X):
        """Store data type as a str so it is easily serializable."""
        return type(X).__name__

    def fit(self, X: Data, sample_dims: Dims, feature_dims: Dims) -> Self:
        """Fit the stacker.

        Parameters
        ----------
        data : DataArray
            The data to be reshaped.

        Returns
        -------
        self : DataArrayStacker
            The fitted stacker.

        """
        self._sanity_check(X, sample_dims, feature_dims)

        self.data_type = self._type_name(X)
        self.sample_dims = sample_dims
        self.feature_dims = feature_dims
        self.dims_mapping.update(
            {
                self.sample_name: sample_dims,
                self.feature_name: feature_dims,
            }
        )

        # Set dimensions and coordinates
        self.dims_in = X.dims
        self.coords_in = {dim: X.coords[dim] for dim in X.dims}

        return self

    def transform(self, X: Data) -> DataArray:
        """Reshape DataArray to 2D.

        Parameters
        ----------
        X : DataArray
            The data to be reshaped.

        Returns
        -------
        DataArray
            The reshaped data.

        Raises
        ------
        ValueError
            If the data to be transformed has different dimensions than the data used to fit the stacker.
        ValueError
            If the data to be transformed has different coordinates than the data used to fit the stacker.

        """
        # Test whether sample and feature dimensions are present in data array
        self._validate_transform_dimensions(X)

        # Check if data to be transformed has the same feature coordinates as the data used to fit the stacker
        self._validate_transform_feature_coords(X)

        # Stack data
        sample_dims = self.dims_mapping[self.sample_name]
        feature_dims = self.dims_mapping[self.feature_name]
        da: DataArray = self._stack(
            X, sample_dims=sample_dims, feature_dims=feature_dims
        )

        # Set out coordinates
        self.coords_out.update(
            {
                self.sample_name: da.coords[self.sample_name],
                self.feature_name: da.coords[self.feature_name],
            }
        )
        return da

    def fit_transform(
        self,
        X: DataVar,
        sample_dims: Dims,
        feature_dims: Dims,
    ) -> DataArray:
        return self.fit(X, sample_dims, feature_dims).transform(X)

    def inverse_transform_data(self, X: DataArray) -> Data:
        """Reshape the 2D data (sample x feature) back into its original dimensions.

        Parameters
        ----------
        X : DataArray
            The data to be reshaped.

        Returns
        -------
        DataArray
            The reshaped data.

        """
        match self.data_type:
            case "DataArray":
                return self._unstack_to_dataarray(X)
            case "Dataset":
                return self._unstack_to_dataset_data(X)
            case _:
                raise TypeError(f"Invalid data type {self._type_name(X)}.")

    def inverse_transform_components(self, X: DataArray) -> Data:
        """Reshape the 2D components (feature x mode) back into its original dimensions.

        Parameters
        ----------
        X : DataArray
            The data to be reshaped.

        Returns
        -------
        DataArray
            The reshaped data.

        """
        match self.data_type:
            case "DataArray":
                return self._unstack_to_dataarray(X)
            case "Dataset":
                return self._unstack_to_dataset_components(X)
            case _:
                raise TypeError(f"Invalid data type {self._type_name(X)}.")

    def inverse_transform_scores(self, X: DataArray) -> DataArray:
        """Reshape the 2D scores (sample x mode) back into its original dimensions.

        Use this for fitted scores.

        Parameters
        ----------
        X : DataArray
            The data to be reshaped.

        Returns
        -------
        DataArray
            The reshaped data.

        """
        return self._unstack_to_dataarray(X)

    def inverse_transform_scores_unseen(self, X: DataArray) -> DataArray:
        """Reshape the 2D scores (sample x mode) back into its original dimensions.

        Use this for new, unseen scores.

        Parameters
        ----------
        X : DataArray
            The data to be reshaped.

        Returns
        -------
        DataArray
            The reshaped data.

        """
        return self.inverse_transform_scores(X)



import pytest

import xeofs as xe

pytest.importorskip("numba")

# =============================================================================
# GENERALLY VALID TEST CASES
# =============================================================================
N_ARRAYS = [1, 2]
N_SAMPLE_DIMS = [1, 2]
N_FEATURE_DIMS = [1, 2]
INDEX_POLICY = ["index"]
NAN_POLICY = ["no_nan", "fulldim"]
DASK_POLICY = ["no_dask"]
SEED = [0]

VALID_TEST_DATA = [
    (na, ns, nf, index, nan, dask)
    for na in N_ARRAYS
    for ns in N_SAMPLE_DIMS
    for nf in N_FEATURE_DIMS
    for index in INDEX_POLICY
    for nan in NAN_POLICY
    for dask in DASK_POLICY
]


# TESTS
# =============================================================================
@pytest.mark.parametrize(
    "kernel",
    [("bisquare"), ("gaussian"), ("exponential")],
)
def test_fit(mock_data_array, kernel):
    gwpca = xe.single.GWPCA(
        n_modes=2, metric="haversine", kernel=kernel, bandwidth=5000
    )
    gwpca.fit(mock_data_array, dim=("lat", "lon"))
    gwpca.components()
    gwpca.largest_locally_weighted_components()


@pytest.mark.parametrize(
    "metric, kernel, bandwidth",
    [
        ("haversine", "invalid_kernel", 5000),
        ("invalid_metric", "gaussian", 5000),
        ("haversine", "exponential", 0),
    ],
)
def test_fit_invalid(mock_data_array, metric, kernel, bandwidth):
    with pytest.raises(ValueError):
        xe.single.GWPCA(n_modes=2, metric=metric, kernel=kernel, bandwidth=bandwidth)
